from tempfile import mkdtemp
import os
import numpy as np
import pytest
from silx.io.url import DataUrl
from silx.third_party.EdfFile import EdfFile
from silx.io.dictdump import dicttoh5
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.preproc.flatfield import FlatFieldDataUrls

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.preproc.flatfield_cuda import CudaFlatFieldDataUrls



@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    cls.n_radios = 10
    cls.n_z = 100
    cls.n_x = 512
    cls.create_radios()
    cls.create_flats_and_dark()
    cls.create_multiple_flats()
    if __has_pycuda__:
        cls.ctx = get_cuda_context()
    yield
    # tear-down
    for fname in ["flat.edf", "dark.edf", "flat00.edf", "flat09.edf"]:
        os.remove(os.path.join(cls.tempdir, fname))
    os.rmdir(cls.tempdir)


@pytest.mark.usefixtures('bootstrap')
class TestFlatField:

    @classmethod
    def create_radios(cls):
        cls.radio_shape = (cls.n_z, cls.n_x)
        cls.radios_stack = np.ones(
            (cls.n_radios, cls.n_z, cls.n_x),
            dtype=np.float32
        )
        for i in range(cls.n_radios):
            cls.radios_stack[i] += i

    @classmethod
    def create_flats_and_dark(cls):
        cls.tempdir = mkdtemp(prefix="nabu_")
        cls.dark = np.ones(cls.radio_shape, dtype=np.float32)
        cls.flat = np.zeros(cls.radio_shape, dtype=np.float32) + 0.5
        # TODO use hdf5 once an official convention is validated
        flat_fname = os.path.join(cls.tempdir, "flat.edf")
        dark_fname = os.path.join(cls.tempdir, "dark.edf")
        edf_flat = EdfFile(flat_fname, "w")
        edf_dark = EdfFile(dark_fname, "w")
        edf_flat.WriteImage({}, cls.flat)
        edf_dark.WriteImage({}, cls.dark)
        cls.flat_url = {0: DataUrl(
            file_path=flat_fname,
            data_path=None, data_slice=[0],
            scheme='fabio'
        )}
        cls.dark_url = {0: DataUrl(
            file_path=dark_fname,
            data_path=None, data_slice=[0],
            scheme='fabio'
        )}

    @classmethod
    def create_multiple_flats(cls):
        flat00 = np.zeros(cls.radio_shape, dtype=np.float32) + 2.
        flat09 = np.zeros_like(flat00) + 11.
        fnames = []
        for fname, arr in zip(["flat00.edf", "flat09.edf"], [flat00, flat09]):
            full_fname = os.path.join(cls.tempdir, fname)
            fnames.append(full_fname)
            edf = EdfFile(full_fname)
            edf.WriteImage({}, arr)
        cls.flats_urls = {
            0: DataUrl(file_path=fnames[0], data_path=None, data_slice=[0], scheme='fabio'),
            9: DataUrl(file_path=fnames[1], data_path=None, data_slice=[0], scheme='fabio')
        }


    def _check_corrected_radios(self, radios_corr, expected_values):
        # must be the same value everywhere in the radio
        std = np.std(np.std(radios_corr, axis=-1), axis=-1)
        assert np.max(np.abs(std)) < 1e-7
        # radios values must be 0, -2, -4, ...
        assert np.allclose(radios_corr[:, 0, 0], expected_values)


    def test_flatfield_simple(self):
        """
        Test the flat-field normalization on a radios stack with 1 dark and 1 flat.

        (I - D)/(F - D)   where I = (1, 2, ...), D = 1, F = 0.5
        = (0, -2, -4, -6, ...)
        """
        flatfield = FlatFieldDataUrls(self.radios_stack.shape, self.flat_url, self.dark_url)
        radios_corr = flatfield.normalize_radios(np.copy(self.radios_stack))
        self._check_corrected_radios(radios_corr, np.arange(0, -2*self.n_radios, -2))


    def test_flatfield_simple_subregion(self):
        """
        Same as test_flatfield_simple, but in a vertical subregion of the radios.
        """
        end_z = 51
        radios_chunk = np.copy(self.radios_stack[:, :end_z, :])
        # we only have a chunk in memory. Instantiate the class with the
        # corresponding subregion to only load the relevant part of dark/flat
        flatfield = FlatFieldDataUrls(
            radios_chunk.shape, self.flat_url, self.dark_url,
            sub_region=(None, None, None, end_z) # start_x, end_x, start_z, end_z
        )
        radios_corr = flatfield.normalize_radios(radios_chunk)
        self._check_corrected_radios(radios_corr, np.arange(0, -2*self.n_radios, -2))


    def test_flatfield_linear_interp(self):
        """
        Test flat-field normalization with 1 dark and 2 flats, with linear
        interpolation between flats.
        I   = 1   2   3   4   5   6   7   8   9   10
        D   = 1                                         (one dark)
        F   = 2                                   11    (two flats)
        F_i = 2   3   4   5   6   7   8   9   10  11   (interpolated flats)
        R   = 0  .5  .66 .75 .8  .83  .86
            = (I-D)/(F-D)
            = (I-1)/I
        """
        flatfield = FlatFieldDataUrls(self.radios_stack.shape, self.flats_urls, self.dark_url)
        radios_corr = flatfield.normalize_radios(np.copy(self.radios_stack))
        I = np.arange(1, self.n_radios+1)
        self._check_corrected_radios(radios_corr, (I-1)/I)

        # Test 2: one of the flats is not at the beginning/end
        # I   = 1    2    3    4    5    6    7    8    9    10
        # F   = 2                       11
        # F_i = 2   3.8  5.6 7.4  9.2   11   11   11   11    11
        # R   = 0   .357  .435 .469 .488 .5  .6   .7  .8     .9
        flats_urls = self.flats_urls.copy()
        flats_urls[5] = flats_urls[9]
        flats_urls.pop(9)
        flatfield = FlatFieldDataUrls(self.radios_stack.shape, flats_urls, self.dark_url)
        radios_corr = flatfield.normalize_radios(np.copy(self.radios_stack))
        I = np.arange(1, self.n_radios+1)
        D = 1
        F = np.array([2, 3.8, 5.6, 7.4, 9.2, 11, 11, 11, 11, 11])
        self._check_corrected_radios(radios_corr, (I-D)/(F-D))


    @pytest.mark.skipif(not(__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_flatfield(self):
        """
        Test the flat-field with cuda back-end.
        """
        d_radios = garray.to_gpu(self.radios_stack.astype("f"))
        cuda_flatfield = CudaFlatFieldDataUrls(
            d_radios.shape,
            self.flats_urls,
            self.dark_url,
            # ~ cuda_options={"ctx": self.ctx}
        )
        cuda_flatfield.normalize_radios(d_radios)
        radios_corr = d_radios.get()
        I = np.arange(1, self.n_radios+1)
        self._check_corrected_radios(radios_corr, (I-1)/I)


# This test should be closer to the ESRF standard setting.
# There are 2 flats, one dark, 4000 radios.
#  dark :  indice=0     value=10
#  flat1 : indice=1     value=4202
#  flat2 : indice=2102  value=2101
#
# The projections have the following indices:
#  j:    0     1        1998  1999  2000  2001                   3999
# idx:  [102, 103, ..., 2100, 2101, 2203, 2204, ..., 4200, 4201, 4202]
#  Notice the gap in the middle.
#
# The linear interpolation is
#    flat_i = (n2 - i)/(n2 - n1)*flat_1  + (i - n1)/(n2 - n1)*flat_2
#  where n1 and n2 are the indices of flat_1 and flat_2 respectively.
# With the above values, we have flat_i = 4203 - i.
#
# The projections values are dark + i*(flat_i - dark),
# so that the normalization norm_i = (proj_i - dark)/(flat_i - dark) gives
#
#   idx     102    103    104    ...
#   flat    4101   4102   4103   ...
#   norm    102    103    104    ...
#


class FlatFieldTestDataset:
    # Parameters
    shp = (27, 32)
    n1 = 1 # flat indice 1
    n2 = 2102 # flat indice 2
    dark_val = 10
    darks = {0: np.zeros(shp, "f") + dark_val}
    flats = {
        n1: np.zeros(shp, "f") + (n2-1)*2,
        n2: np.zeros(shp, "f") + n2-1
    }
    projs_idx = list(range(102, 2102)) + list(range(2203, 4203)) # gap in the middle


    def __init__(self):
        self._generate_projections()
        self._dump_to_h5()
        self._generate_dataurls()


    def get_flat_idx(self, proj_idx):
        flats_idx = sorted(list(self.flats.keys()))
        if proj_idx <= flats_idx[0]:
            return (flats_idx[0], )
        elif proj_idx > flats_idx[0] and proj_idx < flats_idx[1]:
            return flats_idx
        else:
            return (flats_idx[1], )

    def get_flat(self, idx):
        flatidx = self.get_flat_idx(idx)
        if len(flatidx) == 1:
            flat = self.flats[flatidx[0]]
        else:
            nf1, nf2 = flatidx
            w1 = (nf2 - idx)/(nf2 - nf1)
            flat = w1 * self.flats[nf1] + (1 - w1) * self.flats[nf2]
        return flat


    def _generate_projections(self):
        self.projs_data = np.zeros((len(self.projs_idx), ) + self.shp, "f")
        self.projs = {}
        for i, proj_idx in enumerate(self.projs_idx):

            flat = self.get_flat(proj_idx)

            proj_val = self.dark_val + proj_idx * (flat[0,0] - self.dark_val)
            self.projs[str(proj_idx)] = np.zeros(self.shp, "f") + proj_val
            self.projs_data[i] = self.projs[str(proj_idx)]


    def _dump_to_h5(self):
        self.tempdir = mkdtemp(prefix="nabu_")
        self.fname = os.path.join(self.tempdir, "projs_flats.h5")
        dicttoh5(
            {
                "projs": self.projs,
                "flats": {str(k): v for k, v in self.flats.items()},
                "darks": {str(k): v for k, v in self.darks.items()}
            },
            h5file=self.fname
        )


    def _generate_dataurls(self):
        self.flats_urls = {}
        for idx in self.flats.keys():
            self.flats_urls[int(idx)] = DataUrl(
                file_path=self.fname, data_path="/flats/%d" % idx
            )
        self.darks_urls = {}
        for idx in self.darks.keys():
            self.darks_urls[int(idx)] = DataUrl(
                file_path=self.fname, data_path="/darks/0"
            )


@pytest.fixture(scope='class')
def bootstraph5(request):
    cls = request.cls

    cls.dataset = FlatFieldTestDataset()
    n1, n2 = cls.dataset.n1, cls.dataset.n2

    # Interpolation function
    cls._weight1 = lambda i: (n2 - i)/(n2 - n1)

    cls.tol = 5e-4
    cls.tol_std = 1e-3

    yield
    # tear-down
    os.remove(cls.dataset.fname)
    os.rmdir(cls.dataset.tempdir)

@pytest.mark.usefixtures('bootstraph5')
class TestFlatFieldH5:

    def check_normalization(self, projs):
        # Check that each projection is filled with the same values
        std_projs = np.std(projs, axis=(-2, -1))
        assert np.max(np.abs(std_projs)) < self.tol_std
        # Check that the normalized radios are equal to 102, 103, 104, ...
        errs = projs[:, 0, 0] - self.dataset.projs_idx
        assert np.max(np.abs(errs)) < self.tol, "Something wrong with flat-field normalization"


    def test_flatfield(self):
        flatfield = FlatFieldDataUrls(
            self.dataset.projs_data.shape,
            self.dataset.flats_urls,
            self.dataset.darks_urls,
            radios_indices=self.dataset.projs_idx,
            interpolation="linear"
        )
        projs = np.copy(self.dataset.projs_data)
        flatfield.normalize_radios(projs)
        self.check_normalization(projs)


    @pytest.mark.skipif(not(__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_flatfield(self):
        d_projs = garray.to_gpu(self.dataset.projs_data)
        cuda_flatfield = CudaFlatFieldDataUrls(
            self.dataset.projs_data.shape,
            self.dataset.flats_urls,
            self.dataset.darks_urls,
            radios_indices=self.dataset.projs_idx
        )
        cuda_flatfield.normalize_radios(d_projs)
        projs = d_projs.get()
        self.check_normalization(projs)




#
# Another test with more than two flats.
#
# Here we have
#
#   F_i = i + 2
#   R_i = i*(F_i - 1) + 1
#   N_i = (R_i - D)/(F_i - D) = i*(F_i - 1)/( F_i - 1) = i
#


def generate_test_flatfield(n_radios, radio_shape, flat_interval, h5_fname):
    radios = np.zeros((n_radios, ) + radio_shape, "f")
    dark_data = np.ones(radios.shape[1:], "f")
    tempdir = mkdtemp(prefix="nabu_")
    testffname = os.path.join(tempdir, h5_fname)
    flats = {}
    flats_urls = {}
    # F_i = i + 2
    # R_i = i*(F_i - 1) + 1
    # N_i = (R_i - D)/(F_i - D) = i*(F_i - 1)/( F_i - 1) = i
    for i in range(n_radios):
        f_i = i + 2
        if (i % flat_interval) == 0:
            flats["flats_%04d" % i] = np.zeros(radio_shape, "f") + f_i
            flats_urls[i] = DataUrl(file_path=testffname, data_path=str("/flats/flats_%04d" % i), scheme="silx")
        radios[i] = i * (f_i - 1) + 1
    dark = {"dark_0000": dark_data}
    dicttoh5(flats, testffname, h5path="/flats", mode="w")
    dicttoh5(dark, testffname, h5path="/dark", mode="a")
    dark_url = {0: DataUrl(file_path=testffname, data_path="/dark/dark_0000", scheme="silx")}
    return radios, flats_urls, dark_url


@pytest.fixture(scope='class')
def bootstrap_multiflats(request):
    cls = request.cls

    n_radios = 50
    radio_shape = (20, 21)
    cls.flat_interval = 11
    h5_fname = "testff.h5"

    radios, flats, dark = generate_test_flatfield(
        n_radios, radio_shape, cls.flat_interval, h5_fname
    )
    cls.radios = radios
    cls.flats_urls = flats
    cls.darks_urls = dark
    cls.expected_results = np.arange(n_radios)

    cls.tol = 5e-4
    cls.tol_std = 1e-4


@pytest.mark.usefixtures('bootstrap_multiflats')
class TestFlatFieldMultiFlat:

    def check_normalization(self, projs):
        # Check that each projection is filled with the same values
        std_projs = np.std(projs, axis=(-2, -1))
        assert np.max(np.abs(std_projs)) < self.tol_std
        # Check that the normalized radios are equal to 0, 1, 2, ...
        stop = (projs.shape[0] // self.flat_interval) * self.flat_interval
        errs = projs[:stop, 0, 0] - self.expected_results[:stop]
        assert np.max(np.abs(errs)) < self.tol, "Something wrong with flat-field normalization"


    def test_flatfield(self):
        flatfield = FlatFieldDataUrls(
            self.radios.shape,
            self.flats_urls,
            self.darks_urls,
            interpolation="linear"
        )
        projs = np.copy(self.radios)
        flatfield.normalize_radios(projs)
        print(projs[:, 0, 0])
        self.check_normalization(projs)

    @pytest.mark.skipif(not(__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_flatfield(self):
        d_projs = garray.to_gpu(self.radios)
        cuda_flatfield = CudaFlatFieldDataUrls(
            self.radios.shape,
            self.flats_urls,
            self.darks_urls,
        )
        cuda_flatfield.normalize_radios(d_projs)
        projs = d_projs.get()
        self.check_normalization(projs)



import warnings
import numpy as np
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
# won't be necessary once h5py >= 3.0 required
from h5py import version as h5py_version
from packaging.version import parse as parse_version
#


# This function might be moved elsewhere
def get_compacted_dataslices(urls, subsampling=None):
    """
    Regroup urls to get the data more efficiently.
    Build a structure mapping files indices to information on
    how to load the data: `{indices_set: data_location}`
    where `data_location` contains contiguous indices.

    Parameters
    -----------
    urls: dict
        Dictionary where the key is an integer and the value is a silx `DataUrl`.
    subsampling: int, optional
        Subsampling factor when reading the frames. If an integer `n` is provided,
        then one frame out of `n` will be read.

    Returns
    --------
    merged_urls: dict
        Dictionary with the same keys as the `urls` parameter, and where the
        values are the corresponding `silx.io.url.DataUrl` with merged data_slice.
    """
    def _convert_to_slice(idx):
        if np.isscalar(idx):
            return slice(idx, idx+1)
        # otherwise, assume already slice object
        return idx

    def is_contiguous_slice(slice1, slice2):
        if np.isscalar(slice1):
            slice1 = slice(slice1, slice1+1)
        if np.isscalar(slice2):
            slice2 = slice(slice2, slice2+1)
        return slice2.start == slice1.stop

    def merge_slices(slice1, slice2):
        return slice(slice1.start, slice2.stop)

    if len(urls) == 0:
        return urls

    sorted_files_indices = sorted(urls.keys())
    idx0 = sorted_files_indices[0]
    first_url = urls[idx0]

    merged_indices = [
        [idx0]
    ]
    data_location = [
        [
            first_url.file_path(),
            first_url.data_path(),
            _convert_to_slice(first_url.data_slice())
        ]
    ]
    pos = 0
    curr_fp, curr_dp, curr_slice = data_location[pos]
    for idx in sorted_files_indices[1:]:
        url = urls[idx]
        next_slice = _convert_to_slice(url.data_slice())
        if (url.file_path() == curr_fp) and (url.data_path() == curr_dp) and is_contiguous_slice(curr_slice, next_slice):
            merged_indices[pos].append(idx)
            merged_slices = merge_slices(curr_slice, next_slice)
            data_location[pos][-1] = merged_slices
            curr_slice = merged_slices
        else: # "jump"
            pos += 1
            merged_indices.append([idx])
            data_location.append([
                url.file_path(), url.data_path(), _convert_to_slice(url.data_slice())
            ])
            curr_fp, curr_dp, curr_slice = data_location[pos]

    # Format result
    res = {}
    for ind, dl in zip(merged_indices, data_location):
        # res[tuple(ind)] = DataUrl(file_path=dl[0], data_path=dl[1], data_slice=dl[2])
        res.update(
            dict.fromkeys(ind, DataUrl(file_path=dl[0], data_path=dl[1], data_slice=dl[2]))
        )

    if subsampling is None or subsampling <= 1:
        return res

    # Subsample
    next_pos = 0
    for idx in sorted_files_indices:
        url = res[idx]
        url_str = str(url)
        ds = url.data_slice()
        res[idx] = DataUrl(
            file_path=url.file_path(), data_path=url.data_path(),
            data_slice=slice(next_pos + ds.start, ds.stop, subsampling)
        )
        n_imgs = ds.stop - (ds.start + next_pos)
        next_pos = abs(-n_imgs % subsampling)

    return res


def get_first_hdf5_entry(fname):
    with HDF5File(fname, "r") as fid:
        entry = list(fid.keys())[0]
    return entry


def hdf5_entry_exists(fname, entry):
    with HDF5File(fname, "r") as fid:
        res = fid.get(entry, None) is not None
    return res

def get_h5_value(fname, h5_path, default_ret=None):
    with HDF5File(fname, "r") as fid:
        try:
            val_ptr = fid[h5_path][()]
        except KeyError:
            val_ptr = default_ret
    return val_ptr


def get_h5_str_value(dataset_ptr):
    """
    Get a HDF5 field which can be bytes or str (depending on h5py version !).
    """
    data = dataset_ptr[()]
    if isinstance(data, str):
        return data
    else:
        return bytes.decode(data)


# won't be necessary once h5py >= 3.0 required
def check_h5py_version(logger=None):
    """
    (h5py, hdf5) < (3.0, 1.10.6) have many issues. For example:
       - https://github.com/silx-kit/silx/issues/3277
       - overwrite data using (2.10, 1.10.4) while the data was written with (3.0, 1.12)
    """
    recommendations = [
        ("h5py", h5py_version.version, "3.0"),
        ("libhdf5", h5py_version.hdf5_version, "1.10.6")
    ]
    for what, current_version, recommended_version in recommendations:
        msg = str("You are using %s %s. Unexpected behaviors can be experienced. Bugs were fixed in more recent versions. Please consider upgrading to %s >= %s. The h5py wheels (.whl files) contain the hdf5 library (libhdf5), so you don't need administrator privileges or another python environment."
            % (what, current_version, what, recommended_version)
        )
        if parse_version(current_version) < parse_version(recommended_version):
            warnings.warn(msg)
            if logger is not None:
                logger.warning(msg)


def create_dict_of_indices(images_stack, images_indices):
    """
    From an image stack with the images indices, create a dictionary where
    each index is the image index, and the value is the corresponding image.

    Parameters
    ----------
    images_stack: numpy.ndarray
        A 3D numpy array in the layout (n_images, n_y, n_x)
    images_indices: array or list of int
        Array containing the indices of images in the stack

    Examples
    --------
    Given a simple array stack:

    >>> images_stack = np.arange(3*4*5).reshape((3,4,5))
    ... images_indices = [2, 7, 1]
    ... create_dict_of_indices(images_stack, images_indices)
    ... # returns {2: array1, 7: array2, 1: array3}
    """
    if images_stack.ndim != 3:
        raise ValueError("Expected a 3D array")
    if len(images_indices) != images_stack.shape[0]:
        raise ValueError("images_stack must have as many images as the length of images_indices")
    res = {}
    for i in range(len(images_indices)):
        res[images_indices[i]] = images_stack[i]
    return res


def convert_dict_values(dic, val_replacements, bytes_tostring=False):
    """
    Modify a dictionary to be able to export it with silx.io.dicttoh5
    """
    modified_dic = {}
    for key, value in dic.items():
        if isinstance(key, int): # np.isscalar ?
            key = str(key)
        if isinstance(value, bytes) and bytes_tostring:
            value = bytes.decode(value.tostring())
        if isinstance(value, dict):
            value = convert_dict_values(value, val_replacements, bytes_tostring=bytes_tostring)
        else:
            if isinstance(value, DataUrl):
                value = value.path()
            elif value.__hash__ is not None and value in val_replacements:
                value = val_replacements[value]
        modified_dic[key] = value
    return modified_dic



from ..pipeline.fullfield.chunked_cuda import CudaChunkedPipeline as CudaFullFieldPipeline, CudaChunkedPipelineLimitedMemory as CudaFullFieldPipelineLimitedMemory
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.resources.app.fullfield_cuda was moved to nabu.pipeline.fullfield.chunked_cuda.",
    func_name="fullfield_cuda_pipeline"
)


import os
import posixpath
from tempfile import mkdtemp
import numpy as np
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from tomoscan.esrf.hdf5scan import ImageKey
from ..utils import check_supported, is_writeable, deprecation_warning
from ..io.writer import NXProcessWriter
from ..io.utils import get_first_hdf5_entry, hdf5_entry_exists, get_h5_str_value, get_compacted_dataslices
from ..thirdparty.tomwer_load_flats_darks import get_flats_frm_process_file, get_darks_frm_process_file
from .logger import LoggerOrPrint


val_to_nxkey = {
    "flats": ImageKey.FLAT_FIELD.value,
    "darks": ImageKey.DARK_FIELD.value,
    "projections": ImageKey.PROJECTION.value,
}


def _extract_entry(data_path):
    return data_path.lstrip("/").split("/")[0]


def replace_h5_entry(data_url, new_entry):
    split_path = data_url.data_path().lstrip("/").split("/")
    split_path[0] = new_entry
    new_data_path = "/".join(split_path)
    return DataUrl(
        file_path=data_url.file_path(),
        data_path=new_data_path,
        data_slice=data_url.data_slice(),
        scheme=data_url.scheme
    )


class NXFlatField:
    """
    A helper class to load flats and darks, or to compute the final ones.

    At ESRF, several darks/flats series are acquired during a scan.
    Each series undergoes a reduction (mean, median, ...) to get a "final" image.
    For example, there is one series of flats in the beginning, and one in the end ;
    thus there are two corresponding "final" flats.
    """

    _loading_modes = [None, "force-load", "force-compute"]

    def __init__(
        self, dataset_info, lookup_files=None, results_url=None,
        force_load_existing_results=False, loading_mode=None, logger=None,
        flats_reduction="mean", darks_reduction="mean",
        need_flats_and_darks=True,
    ):
        """
        Initialize a FlatFieldLoader helper.

        Parameters
        -----------
        dataset_info: DatasetAnalyzer
            A DatasetAnalyzer object
        lookup_files: list of DataUrl, optional
            List of paths (DataUrl) to inspect to load existing "final" flats/darks.
            If something is found one of these URL, the data will be loaded from there.
        results_url: silx.io.url.DataUrl, optional
            Url to where to write the results.
            Mind the difference with `lookup_files`: this parameter is for writing.
            If the file already exists, its data will be overwritten !
            If set to None, the results will not be written anywhere.
        force_load_existing_results: bool, optional
            DEPRECATED - please use the 'loading_mode' parameter instead.
            Whether to force loading existing results, regardless of the input data
            and results file.
            If set to True, the parameter 'lookup_files' must contain exactly one
            (valid) DataUrl of the results to load.
            Default is False.
        loading_mode: str, optional
            Which approach to take for generating the reduced flats/darks:

               - None (default). First attempt to load already-computed flats/darks from a file,
                 if not present, do the computations.
               - force-load: Load the flats/darks from a given file.
               - force-compute: Discard any existing file and perform the computations
        logger: Logger object, optional
            Logging object
        flats_reduction: str, optional
            Reduction function to use for flats. Can be "mean", "median" or "sum".
            Default is "mean".
        darks_reduction: str, optional
            Reduction function to use for darks. Can be "mean", "median" or "sum".
            Default is "mean".
        need_flats_and_darks: bool, optional
            Whether both flats and darks are needed to compute the final darks/flats.
            If set to True and either no flat/dark is found, an error will be raised.
            Default is True.


        Warning
        -------
        Make sure to use DataUrl objects for the 'lookup_files' and 'results_url'
        parameters.
        """
        if dataset_info.kind != "hdf5":
            raise NotImplementedError("Unsupported dataset type")
        self.dataset_info = dataset_info
        self.data_url = dataset_info.dataset_hdf5_url
        self.image_keys = dataset_info.dataset_scanner.image_key
        self.lookup_files = lookup_files or []
        self.results_url = results_url
        self._set_loading_mode(loading_mode, force_load_existing_results)
        self._need_flats_and_darks = need_flats_and_darks
        self.logger = LoggerOrPrint(logger)
        self.reduction_function = {}
        self._set_reduction_method("flats", flats_reduction)
        self._set_reduction_method("darks", darks_reduction)

        self._get_data_shape()
        self._discover_existing_results()


    def _set_loading_mode(self, mode, force_load_existing_results):
        self.force_load_existing_results = force_load_existing_results
        if force_load_existing_results:
            deprecation_warning(
                "'force_load_existing_results' is replaced with 'mode'",
                func_name="NXFlatField"
            )
        if mode is True or mode is False:
            mode = None
        check_supported(mode, self._loading_modes, "Flatfield loading mode")
        self.loading_mode = mode
        if self.loading_mode == "force-load":
            self.force_load_existing_results = True


    def _set_reduction_method(self, what, reduction_method):
        check_supported(what, val_to_nxkey.keys(), "image type")
        red_methods = {
            "mean": np.mean,
            "median": np.median,
            "sum": np.sum
        }
        check_supported(reduction_method, red_methods, "reduction method")
        self.reduction_function[what] = red_methods[reduction_method]


    def _get_data_shape(self):
        with HDF5File(self.data_url.file_path(), "r") as fid:
            shp = fid[self.data_url.data_path()].shape
        self.data_shape = shp


    def _discover_existing_results(self):
        self._existing_results = None
        if self.loading_mode == "force-compute" or self.lookup_files == []:
            return
        for data_url in self.lookup_files:
            if self.force_load_existing_results or self.is_valid_results_file(data_url):
                self._existing_results = self._get_existing_results_url(data_url)
                self.logger.info("Loaded flats/darks from %s" % data_url.file_path())
                break
        if self._existing_results is None:
            self.logger.debug("Flats/darks could not be loaded from any file")


    def get_config(self):
        if self.results_url is None:
            results_file = "None"
        else:
            results_file = self.results_url.file_path()
        return {
            "input_file": self.data_url.path(),
            "lookup_files": [url.path() for url in self.lookup_files],
            "results_file": results_file,
            "flats_reduction_method": self.reduction_function["flats"].__name__,
            "darks_reduction_method": self.reduction_function["darks"].__name__,
            "image_key": self.image_keys,
            "data_shape": self.data_shape,
        }


    def _is_same_configuration(self, cfg, ignore_filenames=False):
        my_config = self.get_config()
        res = True
        try:
            for key in ["flats_reduction_method", "darks_reduction_method"]:
                res &= (my_config[key] == get_h5_str_value(cfg[key]))

            res &= np.allclose(my_config["data_shape"], cfg["data_shape"][()])
            res &= np.allclose(my_config["image_key"], cfg["image_key"][()])
            du1 = DataUrl(path=my_config["input_file"])
            du2 = DataUrl(path=get_h5_str_value(cfg["input_file"]))
            if not(ignore_filenames):
                res &= (
                    os.path.basename(du1.file_path()) == os.path.basename(du2.file_path())
                )
            res &= (du1.data_path() == du2.data_path())
        except KeyError:
            res = False
        return res


    def is_valid_results_file(self, data_url):
        if not(os.path.isfile(data_url.file_path())):
            return False
        cfg_path = os.path.join(data_url.data_path(), "configuration")
        with HDF5File(data_url.file_path(), "r") as fid:
            try:
                cfg = fid[cfg_path]
            except KeyError:
                return False
            is_same_cfg = self._is_same_configuration(cfg)
            if not(is_same_cfg):
                self.logger.warning("Incompatible results file: %s" % data_url.file_path())
        return is_same_cfg


    def _get_existing_results_url(self, data_url):
        res = {"flats": {}, "darks": {}}
        # If entry is incorrect, take first entry
        # TODO find a more elegant mechanism
        fname = data_url.file_path()
        entry = _extract_entry(data_url.data_path())
        if not(hdf5_entry_exists(fname, entry)):
            new_entry = get_first_hdf5_entry(fname)
            data_url = replace_h5_entry(data_url, new_entry)
            self.logger.warning(
                "Entry %s in file %s does not exist. Using the first entry %s"
                % (entry, fname, new_entry)
            )
        #
        results_path = os.path.join(data_url.data_path(), "results") # NXProcess
        with HDF5File(data_url.file_path(), "r") as fid:
            for what in res.keys():
                for img_id in fid[results_path][what].keys():
                    res[what][int(img_id)] = DataUrl(
                        file_path=data_url.file_path(),
                        data_path=os.path.join(results_path, what, img_id)
                    )
        return res


    @staticmethod
    def load_data(ff_results):
        res_arrays = {"flats": {}, "darks": {}}
        for what in res_arrays.keys():
            for img_id, data_url in ff_results[what].items():
                with HDF5File(data_url.file_path(), "r") as fid:
                    res_arrays[what][img_id] = fid[data_url.data_path()][()]
        return res_arrays


    def get_data_chunk(self, data_slice):
        """
        Get a data chunk from the data volume whose path was specified at class instantiation.
        """
        with HDF5File(self.data_url.file_path(), "r", swmr=True) as f:
            data_raw = f[self.data_url.data_path()][data_slice]
        return data_raw


    def write_results(self, flats_and_darks, process_name="flat_field_images"):
        """
        Write the results in a HDF5 file.

        Parameters
        -----------
        flats_and_darks: dict
            Dictionary of dictionaries, in the form

            >>> {
            ...     "darks": { 0: reduced_flat_0, 1000: reduced_flat_1000},
            ...     "flats": { 1: reduced_dark_0},
            ... }

            where each reduced_array is a numpy array.

        process_name: str, optional
            Process name. Default is "flat_field_images".
        """
        if self.results_url is None:
            return
        results_file = self.results_url.file_path()
        writer = NXProcessWriter(
            results_file,
            entry=_extract_entry(self.results_url.data_path()),
            overwrite=True
        )
        h5_results_path = writer.write(
            flats_and_darks, process_name, config=self.get_config()
        )
        if "flats" in flats_and_darks:
            self.dataset_info.dataset_scanner.save_reduced_flats(flats_and_darks["flats"])
        if "darks" in flats_and_darks:
            self.dataset_info.dataset_scanner.save_reduced_darks(flats_and_darks["darks"])
        self.logger.info("Wrote final flats and darks to %s" % results_file)
        links = {}
        for img_type, imgs in flats_and_darks.items():
            links[img_type] = {}
            for img_idx in imgs.keys():
                links[img_type][img_idx] = DataUrl(
                    file_path=results_file,
                    data_path=os.path.join(h5_results_path, img_type, str(img_idx))
                )
        return links


    def compute_final_images(self):
        self.logger.info(
            "Computing final flats/darks for %s" % self.data_url.file_path()
        )
        res = {"flats": {}, "darks": {}}
        for what in res.keys():
            img_slices = self.dataset_info.get_data_slices(what)
            if img_slices == []:
                err_msg = "No %s found in %s" % (what, self.data_url.file_path())
                self.logger.error(err_msg)
                if self._need_flats_and_darks:
                    raise ValueError(err_msg)
                res[what] = None
            for data_slice in img_slices:
                data_chunk = self.get_data_chunk(data_slice)
                img = self.reduction_function[what](data_chunk, axis=0)
                res[what][int(data_slice.start)] = img
        return res


    def get_final_urls(self):
        if self._existing_results is not None:
            return self._existing_results
        else:
            if self.results_url is None:
                raise ValueError("Cannot call get_final_urls when results do not exist and write_results=None")
            flats_and_darks = self.compute_final_images()
            urls = self.write_results(flats_and_darks)
            return urls


def update_dataset_info_flats_darks(dataset_info, processes_file=None, output_dir=None, flatfield_mode=None):
    """
    Update a DatasetAnalyzer instance with flats/darks images.

    Parameters
    ----------
    dataset_info: DatasetAnalyzer
        DatasetAnalyzer instance
    processes_file: str, optional
        File name of a 'processes.h5' HDF5 file containing final darks/flats to use.
        If the file exists and contains the data, then darks/flats from this file will be used.
    output_dir: str, optional
        Directory where 'processes.h5' will be stored if we can't write in the same directory
        as the dataset.
    flatfield_mode: bool or str, optional
        How to perform flatfield:

           - False indicates that flatfield normalization is entirely disabled
           - True will do the flats/darks reduction by firt atttempting to load results from files
           - "force-compute" will discard files containing existing results and do the reduction
             (so the current dataset has to contain darks/flats)
           - "force-load" will load the reduced flats/darks regardless of the current dataset
             actually containing any flat/dark.
    """
    def _load_flats_from_tomwer():
        tomwer_processes_file = _get_tomwer_process_file_name()
        if tomwer_processes_file is None or not(os.path.isfile(tomwer_processes_file)):
            # if not found try with the old one
            tomwer_processes_file = os.path.join(
                dataset_info.dataset_scanner.path,
                "tomwer_processes.h5"
            )

        # try to load dark and flats from the default url (flats.hdf5 and darks.hdf5 for HDF5TomoScan and dark.edf, flatHSTXXX.edf for EDFTomoScan)
        new_flats = dataset_info.dataset_scanner.load_reduced_flats(return_as_url=True)
        new_darks = dataset_info.dataset_scanner.load_reduced_darks(return_as_url=True)

        if len(new_flats) > 0 and len(new_darks) > 0:
            files_used = str([x.file_path() for x in list(new_darks.values()) + list(new_flats.values())])
        else:
            if tomwer_processes_file is None or not (os.path.isfile(tomwer_processes_file)):
                return False
            new_flats = get_flats_frm_process_file(
                tomwer_processes_file, dataset_info.dataset_scanner.entry
            )
            new_darks = get_flats_frm_process_file(
                tomwer_processes_file, dataset_info.dataset_scanner.entry
            )
            files_used = tomwer_processes_file

        if not (len(new_flats) > 0 and len(new_darks) > 0):
            return False

        dataset_info.logger.info("Loading darks and refs from %s" % files_used)
        dataset_info.flats = new_flats
        dataset_info.darks = new_darks
        return True


    def _get_results_file():
        if processes_file not in (None, ""):
            if is_writeable(os.path.dirname(processes_file)):
                return processes_file
            else:
                dataset_info.logger.error("Cannot write to %s" % processes_file)
        results_relfname = os.path.splitext(
            os.path.basename(dataset_info.dataset_scanner.master_file)
        )[0] + "_nabu_processes.hdf5"
        # Attempt 1: write in the same dir as the master NX file
        results_dir = os.path.dirname(dataset_info.dataset_scanner.master_file)
        if is_writeable(results_dir):
            return os.path.join(results_dir, results_relfname)
        # Attempt 2: write in the "output" dir, if specified
        out_dir = output_dir
        if out_dir is not None and is_writeable(out_dir):
            return os.path.join(out_dir, results_relfname)
        # Last attempt: write in a temporary directory
        tempdir = mkdtemp(prefix="nabu_flatfield_")
        return os.path.join(tempdir, results_relfname)


    def _compute_or_load_flats():
        h5_entry = dataset_info.hdf5_entry or "entry"
        h5_path = posixpath.join(h5_entry, "flat_field_images")
        # In this case, we always want to return a dict of DataUrl,
        # so we have to write the final flats/darks somewhere.
        # Therefore results_url cannot be None when instantiating NXFlatField.
        # It will be ignored if data can be loaded from processes_file
        results_file = _get_results_file()
        results_url = DataUrl(file_path=results_file, data_path=h5_path)

        lookup_files = None
        if processes_file is not None:
            lookup_files = [DataUrl(file_path=processes_file, data_path=h5_path)]
        else:
            if os.path.isfile(results_file):
                # We use an existing "results file"  as lookup file.
                # NXFlatField wont overwrite it if it has the good configuration
                lookup_files = [DataUrl(file_path=results_file, data_path=h5_path)]

        nxflatfield = NXFlatField(
            dataset_info,
            lookup_files=lookup_files,
            results_url=results_url,
            loading_mode=flatfield_mode,
            flats_reduction="median",
            darks_reduction="mean",
            logger=dataset_info.logger
        )
        res = nxflatfield.get_final_urls()
        dataset_info.flats = res["flats"]
        dataset_info.darks = res["darks"]


    def _get_tomwer_process_file_name():
        if dataset_info.dataset_scanner.path is not None:
            basename, _ = os.path.splitext(dataset_info.dataset_scanner.master_file)
            basename = os.path.basename(basename)
            basename = "_".join((basename, "tomwer_processes.h5"))
            return os.path.join(dataset_info.dataset_scanner.path, basename)
        else:
            return None


    def _get_flats_darks():
        if flatfield_mode is False:
            # Flat-fielding is disabled by user
            return
        if len(dataset_info.flats) == 0 and flatfield_mode != "force-load":
            # No flats at all in the dataset. Do nothing.
            return
        if processes_file is None and _load_flats_from_tomwer():
            # Loaded from [XXX]_tomwer_processes.h5
            return
        # Otherwise load or compute flats/darks with nabu
        _compute_or_load_flats()

    _get_flats_darks()




import math
import numpy as np
from ..resources.logger import LoggerOrPrint
from ..misc import fourier_filters
from ..misc.padding import pad_interpolate, recut


class GeoPars:
    """
    A class to describe the geometry of a phase contrast radiography with
    a source obtained by a focussing system, possibly astigmatic, which is
    at distance z1_vh from the sample. The detector is at z2 from the sample
    """

    def __init__(
        self,
        z1_vh=None,
        z2=None,
        pix_size_det=None,
        wavelength=None,
        magnification=True,
        length_scale=10.0e-6,
        logger=None,
    ):
        """
        Parameters
        ----------
        z1_vh : None, a float, or a sequence of two floats
           the source sample distance (meters), if None the parallel beam is assumed.
           If two floats are given then they are taken as the distance of the vertically focused source (horizontal line)
           and the horizontaly focused source (vertical line) for KB mirrors.
        z2 : float
           the sample detector distance (meters).
        pix_size_det: float
           pixel size (meters)
        wavelength: float
           beam wave length (meters).
        magnification: boolean defaults to True
           if false no magnification is considered
        length_scale: float
            rescaling length scale, meant to avoid having too big or too small numbers.
            defaults to 10.0e-6
        logger: Logger, optional
            A logging object
        """
        self.logger = LoggerOrPrint(logger)
        if z1_vh is None:
            self.z1_vh = None
        else:
            if hasattr(type(z1_vh), "__iter__"):
                self.z1_vh = np.array(z1_vh)
            else:
                self.z1_vh = np.array([z1_vh, z1_vh])
        self.z2 = z2
        self.magnification = magnification
        self.pix_size_det = pix_size_det

        if self.magnification and self.z1_vh is not None:
            self.M_vh = (self.z1_vh + self.z2) / self.z1_vh
        else:
            self.M_vh = np.array([1, 1])

        self.logger.debug("Magnification : h ({}) ;  v ({}) ".format(self.M_vh[1], self.M_vh[0]))

        self.length_scale = length_scale
        self.wavelength = wavelength

        self.maxM = self.M_vh.max()

        self.pix_size_rec = self.pix_size_det / self.maxM  # we bring everything to highest magnification

        which_unit = np.sum(np.array([(self.pix_size_rec > small) for small in [1.0e-6, 1.0e-7]]).astype(np.int32))
        self.pixelsize_string = [
            "{:.1f} nm".format(self.pix_size_rec * 1e9),
            "{:.3f} um".format(self.pix_size_rec * 1e6),
            "{:.1f} um".format(self.pix_size_rec * 1e6),
        ][which_unit]

        if self.magnification:
            self.logger.debug(
                "All images are resampled to smallest pixelsize: {}".format(self.pixelsize_string),
            )
        else:
            self.logger.debug("Pixelsize images:  {}".format(self.pixelsize_string))


class CTFPhaseRetrieval:
    def __init__(self, geo_pars, delta_beta, lim1=1.0e-5, lim2=0.2, use_rfft=False, logger=None):
        """
        This class implements the CTF formula (see equation 8 in Boliang Yu et al
        Optic Express Vol 26, No 9, 2018) in its regularised form which avoids the zeros
        of unregularised_filter_denominator
        (unreg_filter_denom is the so here named denominator/delta_beta of equation 8)

        Parameters
        ----------
        geo_pars: GeoPars
            the geometry description
        delta_beta : float
            the delta/beta ratio
        lim1: float >0
            the regulariser strenght at low  frequencies
        lim2: float >0
            the regulariser strenght at high frequencies
        use_rfft: bool, optional
            Whether to use real-to-complex (R2C) FFT instead of usual complex-to-complex (C2C).
        logger: optional
            a logger object
        """
        self.logger = LoggerOrPrint(logger)
        self.setup(geo_pars, delta_beta, lim1=lim1, lim2=lim2, use_rfft=use_rfft)

    def setup(self, geo_pars, delta_beta, lim1=None, lim2=None, use_rfft=False):
        """
        Set up the current geometrical parameters.

        Parameters
        -----------
        geo_pars: GeoPars
            The eometry of the experiment
        delta_beta : float,
            The delta/beta ratio in single-distance phase retrieval
        lim1: float >0
            The regulariser strength at low frequencies
        lim2: float >0
            The regulariser strength at high frequencies
        """
        if not isinstance(geo_pars, GeoPars):
            raise ValueError("Expected GeoPars instance for 'geo_pars' parameter")

        self.lim = None
        # these parameters will be stored for pipelining: if the shape changes in a call
        # to retrieve_phase the filter will be recalculated for that shape
        self.lim1 = lim1
        self.lim2 = lim2
        self.delta_beta = delta_beta
        self.geo_pars = geo_pars
        self.filter_is_ready = False
        self.filter_shape = None
        self.use_rfft = use_rfft
        self._fft_func = np.fft.rfft2 if use_rfft else np.fft.fft2
        self._ifft_func = np.fft.irfft2 if use_rfft else np.fft.ifft2


    def require_filter_for_shape(self, padded_img_shape):
        if self.filter_is_ready and self.filter_shape == padded_img_shape:
            return
        else:
            self.filter_is_ready = False
            self.filter_shape = None

        ## the parameter "length_scale" was mentioned, in the octave code,
        ## as a rescaling length scale, which is meant to avoid
        ## having too big or too small numbers.
        ## From the mathematical point of view, it is infact completely trasparent:
        ## its action is on fsamplex, fsampley and  betash, betasv.
        ## But these latters ( beta's)  are multiplied by  the formers (fsample's) so that
        ## "length_scale" mathematically disappears, however in case of simple precision float
        ## the exponent of a float number ranges from -38 to 38, and one could approach it
        ## as an example by taking the square of a very small number ( 1.0e-19), and losing significant bits
        ## in he mantissa or getting zero, or the square of a big number, thus generting inf
        ## Althought the values involved in our x-ray regimes seems safe, with respect to these problems,
        ## this length_scale parameters does not hurt.
        fsample_vh = np.array(
            [self.geo_pars.length_scale / self.geo_pars.pix_size_rec, self.geo_pars.length_scale / self.geo_pars.pix_size_rec]
        )

        if not self.use_rfft:
            ff_index_vh = list(map(np.fft.fftfreq, padded_img_shape))
        else:
            ff_index_vh = [np.fft.fftfreq(padded_img_shape[0]), np.fft.rfftfreq(padded_img_shape[1])]

        # if padded_img_shape[1]%2 == 0 : # change to holotomo_slave indexing (by a transparent 2pi shift)
        #     ff_index_x[ ff_index_x  == -0.5    ] =  +0.5
        # if padded_img_shape[0]%2 == 0 : # change to holotomo_slave indexing (by a transparent 2pi shift)
        #     ff_index_y[ ff_index_y  == -0.5    ] =  +0.5

        frequencies_vh = np.array(
            np.meshgrid(ff_index_vh[0] * fsample_vh[0], ff_index_vh[1] * fsample_vh[1], indexing="ij")
        )

        frequencies_squared_vh = frequencies_vh * frequencies_vh

        ## ---------------  fresnelnumbers and forward propagators -------------------
        ## In the limit of parallel beam, z1_h and z1_v would be infinite
        ## so that the here below distances becomes z2
        ## which is sample-detector distance.
        ##

        if self.geo_pars.z1_vh is not None:
            distances_vh = (self.geo_pars.z1_vh * self.geo_pars.z2) / (self.geo_pars.z1_vh + self.geo_pars.z2)
        else:
            distances_vh = np.array([self.geo_pars.z2, self.geo_pars.z2])

        ## Citing David Paganin (2002) :
        ## The intensity I_{R_1}(r⊥,z) at a distance z  of a weakly refracting object illuminated by a
        ## point source at distance R1  behind the said object, is related to the intensity
        ## I∞(r⊥,z), which would result from normally
        ## incident collimated illumination of the same object, by (Pogany et al., 1997):
        ##   I_{R_1}(r⊥,z) = 1/{M^2} I∞(r⊥/M , z/M)
        ## where M is the magnification.
        ## This explains the effective distance formula expressed by
        ## distancesh, distancesv  above.
        ## ------------------------------------------------------------------------------

        lambda_dist_vh = self.geo_pars.wavelength * distances_vh / (self.geo_pars.length_scale ** 2)

        ## ---------------------------------------------------------------------------------------
        ## -> cut_v at first maximum of ctf largest distance
        ##
        ## In the paraxial expansion of the Fresnel propagator, the phase is equal to
        ##                   1/2 K_{parallel}^2 wavelength * Distance /2 / pi
        ##
        ## When this is equal to a multiple of 2 pi, the effect of propagation disappears
        ## and we have a singularity in equation 8.
        ##
        ## The sampling in the reciprocal space is done with  a step length of 2*pi*fsamplex,y
        ##   (note: fsamples are the plain inverse of pixel size)
        ## The first singularity occurs  at a frequence number
        ##                 K_{parallel}/2/pi/ fsamplex
        ##
        ##  for K_{parallel}^2 = 2 * (2pi)^2 /( wavelength * distance  )
        ##
        ##  which would correspond to
        ##            K_{parallel}/2/pi/ fsamplex  =  sqrt( 2/wavelength*distance) / fsample
        ##
        ## Question:( why the factor 2 appear at the denominator in the square root below?)
        ## Answer : the below defined cut corresponds to the first maximum of the denominator,
        ## before arriving to the first zero. In this way the regularisation
        ## is already at ( almost) full strenght on the first pole.
        self.cut_v = math.sqrt(1.0 / 2 / lambda_dist_vh[0]) / fsample_vh[0]
        self.cut_v = min(self.cut_v, 0.5)

        self.logger.debug("Normalized cut-off = {:5.3f}".format(self.cut_v))

        self.r = fourier_filters.get_lowpass_filter(
            padded_img_shape,
            cutoff_par=(
                0.5 / (self.cut_v + 1.0 / padded_img_shape[0]),
                0.01 * padded_img_shape[0] / (1 + self.cut_v * padded_img_shape[0]),
            ),
            use_rfft=self.use_rfft
        )
        self.r /= self.r[0, 0]

        self.lim = self.lim1 * self.r + self.lim2 * (1 - self.r)

        # more methods exist in the original code, and they are initialized starting from here
        # (ht_app1, ht_app2... ht_app7)

        fresnel_phase = (
            np.pi * lambda_dist_vh[1] * frequencies_squared_vh[1] + np.pi * lambda_dist_vh[0] * frequencies_squared_vh[0]
        )

        if self.delta_beta:
            self.logger.debug(
                "Taking into account absorption assuming homogeneous object.\n"
                "Set delta_beta = 0 if this is not the purpose."
            )
            unreg_filter_denom = np.sin(fresnel_phase) + (1.0 / self.delta_beta) * np.cos(fresnel_phase)
        else:
            unreg_filter_denom = np.sin(fresnel_phase)

        self.unreg_filter_denom = unreg_filter_denom

        self.filter_is_ready = True
        self.filter_shape = padded_img_shape

    def _apply_filter(self, img):
        self.require_filter_for_shape(img.shape)

        img_f = self._fft_func(img)
        firec0 = self.unreg_filter_denom * img_f
        unreg_filter_denom_0_mean = self.unreg_filter_denom[0, 0]
        nf, mf = img.shape

        # here it is assumed that the average of img is 1 and the DC component is removed
        firec0[0, 0] -= nf * mf * unreg_filter_denom_0_mean

        ## formula 8, with regularisation to stay at a safe distance from the poles
        ph = firec0 / (2 * self.unreg_filter_denom * self.unreg_filter_denom + self.lim)
        ph = self._ifft_func(ph).real
        return ph


    def retrieve_phase(self, img, padded_shape="auto", padding_mode="reflect", translation_vh=None, normalize_by_mean=False):
        """
        Apply the CTF filter to retrieve the phase.

        Parameters
        ----------
        img: np.ndarray
            Projection image. It must have been already flat-fielded.
        padded_shape: str or tuple, optional
            Padded image shape, in the form (num_rows, num_columns) i.e (vertical, horizontal).
            By default, it is twice the image shape.
        padding_mode: str
            Padding mode. It must be valid for the numpy.pad function
        translation_vh: array, optional
            Shift in the form (y, x). It is used to perform a translation of the image before applying the CTF filter.
        normalize_by_mean: bool
            Whether to divide the (padded) image with its mean before applying the CTF filter.

        Returns
        --------
        ph: numpy.ndarray
            Phase image
        """
        if padded_shape is None or padded_shape is False:
            padded_shape = img.shape # no padding
        elif isinstance(padded_shape, (tuple, list, np.ndarray)):
            pass
        elif padded_shape == "auto":
            padded_shape = (2 * img.shape[0], 2 * img.shape[1])

        padded_img = pad_interpolate(
            img, padded_shape, translation_vh=translation_vh, padding_mode=padding_mode
        )
        if normalize_by_mean:
            padded_img /= padded_img.mean()
        phase_img = self._apply_filter(padded_img)
        res = recut(phase_img, img.shape)
        return res


CtfFilter = CTFPhaseRetrieval


# COMPAT.
from ..utils import deprecated_class
from ..reconstruction.rings_cuda import CudaMunchDeringer as CudaMunchDeringer_

CudaMunchDeringer = deprecated_class(
    "preproc.rings was moved to reconstruction.rings",
    do_print=True
)(CudaMunchDeringer_)



import numpy as np
from ..utils import get_cuda_srcfile, updiv
from ..cuda.utils import __has_pycuda__
from .histogram import PartialHistogram, VolumeHistogram

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from ..cuda.kernel import CudaKernel
    from ..cuda.processing import CudaProcessing


class CudaPartialHistogram(PartialHistogram):
    def __init__(
        self,
        method="fixed_bins_number",
        bin_width="uint16",
        num_bins=None,
        min_bins=None,
        cuda_options=None,
    ):
        if method == "fixed_bins_width":
            raise NotImplementedError(
                "Histogram with fixed bins width is not implemented with the Cuda backend"
            )
        super().__init__(
            method=method,
            bin_width=bin_width,
            num_bins=num_bins,
            min_bins=min_bins,
        )
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._init_cuda_histogram()


    def _init_cuda_histogram(self):
        self.cuda_hist = CudaKernel(
            "histogram",
            filename=get_cuda_srcfile("histogram.cu"),
            signature="PiiiffPi",
        )
        self.d_hist = garray.zeros(self.num_bins, dtype=np.uint32)

    def _compute_histogram_fixed_nbins(self, data, data_range=None):
        if isinstance(data, np.ndarray):
            data = garray.to_gpu(data)
        if data_range is None:
            # Should be possible to do both in one single pass with ReductionKernel
            # and garray.vec.float2, but the last step in volatile shared memory
            # still gives errors. To be investigated...
            data_min = garray.min(data).get()[()]
            data_max = garray.max(data).get()[()]
        else:
            data_min, data_max = data_range
        Nz, Ny, Nx = data.shape
        block = (16, 16, 4)
        grid = (
            updiv(Nx, block[0]),
            updiv(Ny, block[1]),
            updiv(Nz, block[2]),
        )
        self.d_hist.fill(0)
        self.cuda_hist(
            data,
            Nx,
            Ny,
            Nz,
            data_min,
            data_max,
            self.d_hist,
            self.num_bins,
            grid=grid,
            block=block,
        )
        # Return a result in the same format as numpy.histogram
        res_hist = self.d_hist.get()
        res_bins = np.linspace(data_min, data_max, num=self.num_bins + 1, endpoint=True)
        return res_hist, res_bins


class CudaVolumeHistogram(VolumeHistogram):
    def __init__(
        self,
        data_url,
        chunk_size_slices=100,
        chunk_size_GB=None,
        nbins=1e6,
        logger=None,
        cuda_options=None,
    ):
        self.cuda_options = cuda_options
        super().__init__(
            data_url,
            chunk_size_slices=chunk_size_slices,
            chunk_size_GB=chunk_size_GB,
            nbins=nbins,
            logger=logger,
        )

    def _init_histogrammer(self):
        self.histogrammer = CudaPartialHistogram(
            method="fixed_bins_number",
            num_bins=self.nbins,
            cuda_options=self.cuda_options,
        )


    def _compute_histogram(self, data):
        return self.histogrammer.compute_histogram(data) # 3D



import numpy as np
import pytest
from scipy.misc import ascent
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.utils import calc_padding_lengths, get_cuda_srcfile
from nabu.testutils import get_data, generate_tests_scenarios
if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.cuda.kernel import CudaKernel
    from nabu.cuda.padding import CudaPadding

scenarios_legacy = [
    {
        "shape": (512, 501),
        "shape_padded": (1023, 1022),
        "constant_values": ((1., 2.), (3., 4.)),
    },
]

# parametrize with fixture and "params=" will launch a new class for each scenario.
# the attributes set to "cls" will remain for all the tests done in this class
# with the current scenario.
@pytest.fixture(scope='class', params=scenarios_legacy)
def bootstrap_legacy(request):
    cls = request.cls
    cls.data = get_data("mri_proj_astra.npz")["data"]
    cls.tol = 1e-7
    cls.params = request.param
    cls.ctx = get_cuda_context()
    cls._calc_pad()
    cls._init_kernels()


@pytest.mark.skipif(not(__has_pycuda__), reason="Need Cuda and pycuda for this test")
@pytest.mark.usefixtures('bootstrap_legacy')
class TestPaddingLegacy:

    @classmethod
    def _calc_pad(cls):
        cls.shape = cls.params["shape"]
        cls.data = np.ascontiguousarray(cls.data[:cls.shape[0], :cls.shape[1]])
        cls.shape_padded = cls.params["shape_padded"]
        ((pt, pb), (pl, pr)) = calc_padding_lengths(cls.shape, cls.shape_padded)
        cls.pad_top_len = pt
        cls.pad_bottom_len = pb
        cls.pad_left_len = pl
        cls.pad_right_len = pr

    @classmethod
    def _init_kernels(cls):
        cls.pad_kern = CudaKernel(
            "padding_constant",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiiiffff",
        )
        cls.pad_edge_kern = CudaKernel(
            "padding_edge",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiii",
        )
        cls.d_data_padded = garray.zeros(cls.shape_padded, "f")

    def _init_padding(self, arr=None):
        arr = arr or self.data
        self.d_data_padded.fill(0)
        Ny, Nx = self.shape
        self.d_data_padded[:Ny, :Nx] = self.data

    def _pad_numpy(self, arr=None, **np_pad_kwargs):
        arr = arr or self.data
        data_padded_ref = np.pad(
            arr,
            ((self.pad_top_len, self.pad_bottom_len), (self.pad_left_len, self.pad_right_len)),
            **np_pad_kwargs
        )
        # Put in the FFT layout
        data_padded_ref = np.roll(
            data_padded_ref,
            (-self.pad_top_len, -self.pad_left_len),
            axis=(0, 1)
        )
        return data_padded_ref

    def test_constant_padding(self):
        self._init_padding()
        # Pad using the cuda kernel
        ((val_top, val_bottom), (val_left, val_right)) = self.params["constant_values"]
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded

        self.pad_kern(
            self.d_data_padded,
            Nx, Ny,
            Nxp, Nyp,
            self.pad_left_len, self.pad_right_len,
            self.pad_top_len, self.pad_bottom_len,
            val_left, val_right, val_top, val_bottom
        )
        # Pad using numpy
        data_padded_ref = self._pad_numpy(
            mode="constant",
            constant_values=self.params["constant_values"]
        )
        # Compare
        errmax = np.max(np.abs(self.d_data_padded.get() - data_padded_ref))
        assert errmax < self.tol, "Max error is too high"

    def test_edge_padding(self):
        self._init_padding()
        # Pad using the cuda kernel
        ((val_top, val_bottom), (val_left, val_right)) = self.params["constant_values"]
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded

        self.pad_edge_kern(
            self.d_data_padded,
            Nx, Ny,
            Nxp, Nyp,
            self.pad_left_len, self.pad_right_len,
            self.pad_top_len, self.pad_bottom_len,
        )
        # Pad using numpy
        data_padded_ref = self._pad_numpy(mode="edge")
        # Compare
        errmax = np.max(np.abs(self.d_data_padded.get() - data_padded_ref))
        assert errmax < self.tol, "Max error is too high"


scenarios = generate_tests_scenarios({
    "shape": [(511, 512), (512, 511)],
    "pad_width": [((256, 255), (128, 127)), ((0, 0), (6, 7))],
    "mode": CudaPadding.supported_modes if __has_pycuda__ else [],
    "constant_values": [0, ((1., 2.), (3., 4.))],
    "output_is_none": [True, False],
})

@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    cls.data = ascent().astype("f")
    cls.tol = 1e-7
    cls.ctx = get_cuda_context()


@pytest.mark.skipif(not(__has_pycuda__), reason="Need Cuda and pycuda for this test")
@pytest.mark.usefixtures('bootstrap')
class TestCudaPadding:

    @pytest.mark.parametrize("config", scenarios)
    def test_padding(self, config):
        shape = config["shape"]
        data = self.data[:shape[0], :shape[1]]
        kwargs = {}
        if config["mode"] == "constant":
            kwargs["constant_values"] = config["constant_values"]
        ref = np.pad(
            data,
            config["pad_width"],
            mode=config["mode"],
            **kwargs
        )
        if config["output_is_none"]:
            output = None
        else:
            output = garray.zeros(ref.shape, "f")
        cuda_padding = CudaPadding(
            config["shape"],
            config["pad_width"],
            mode=config["mode"],
            constant_values=config["constant_values"],
            cuda_options={"ctx": self.ctx}
        )
        d_img = garray.to_gpu(np.ascontiguousarray(data, dtype="f"))
        res = cuda_padding.pad(d_img, output=output)

        err_max = np.max(np.abs(res.get() - ref))
        assert err_max < self.tol, str("Something wrong with padding for configuration %s" % (str(config)))


    def test_custom_coordinate_transform(self):
        data = self.data
        R, C = np.indices(data.shape, dtype=np.int32)

        pad_width = ((256, 255), (254, 251))
        mode = "reflect"

        coords_R = np.pad(R, pad_width, mode=mode)
        coords_C = np.pad(C, pad_width, mode=mode)
        # Further transform of coordinates - here FFT layout
        coords_R = np.roll(coords_R, (-pad_width[0][0], -pad_width[1][0]), axis=(0, 1))
        coords_C = np.roll(coords_C, (-pad_width[0][0], -pad_width[1][0]), axis=(0, 1))

        cuda_padding = CudaPadding(
            data.shape,
            (coords_R, coords_C),
            mode=mode,
        )
        d_img = garray.to_gpu(data)
        d_out = garray.zeros(cuda_padding.padded_shape, "f")
        res = cuda_padding.pad(d_img, output=d_out)

        ref = np.roll(
            np.pad(data, pad_width, mode=mode),
             (-pad_width[0][0], -pad_width[1][0]),
             axis=(0, 1)
        )

        err_max = np.max(np.abs(d_out.get() - ref))
        assert err_max < self.tol, "Something wrong with custom padding"






import numpy as np
import pytest
from nabu.testutils import get_big_data, __big_testdata_dir__, compare_arrays, generate_tests_scenarios, __do_long_tests__
from nabu.cuda.utils import __has_pycuda__, __has_cufft__, get_cuda_context
__has_cuda_fbp__ = (__has_cufft__ and __has_pycuda__)
if __has_cuda_fbp__:
    from nabu.reconstruction.reconstructor_cuda import CudaReconstructor
    from nabu.reconstruction.fbp import Backprojector as CudaBackprojector
    import pycuda.gpuarray as garray





scenarios = generate_tests_scenarios({
    "axis": ["z", "y", "x"],
    "vol_type": ["sinograms", "projections"],
    "indices": [(300, 310)], # reconstruct 10 slices
    "slices_roi": [
        None,
        (None, None, 200, 400),
        (250, 300, None, None),
        (120, 330, 301, 512)
    ],
})


@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    cls.projs = get_big_data("MRI512_projs.npy")
    if __has_cuda_fbp__:
        cls.ctx = get_cuda_context()
        cls.d_projs = garray.to_gpu(cls.projs)
    cls.ref = None
    cls.tol = 5e-2


@pytest.mark.skipif(
    __big_testdata_dir__ is None or not(__do_long_tests__),
    reason="need environment variable NABU_BIGDATA_DIR and NABU_LONG_TESTS=1"
)
@pytest.mark.usefixtures('bootstrap')
class TestReconstructor:

    @pytest.mark.skipif(not(__has_cuda_fbp__), reason="need pycuda and scikit-cuda")
    @pytest.mark.parametrize("config", scenarios)
    def test_cuda_reconstructor(self, config):
        data = self.projs
        d_data = self.d_projs
        if config["vol_type"] == "sinograms":
            data = np.moveaxis(self.projs, 1, 0)
            d_data = self.d_projs.transpose(axes=(1, 0, 2)) # view
        reconstructor = CudaReconstructor(
            data.shape,
            config["indices"],
            axis=config["axis"],
            vol_type=config["vol_type"],
            slices_roi=config["slices_roi"]
        )
        res = reconstructor.reconstruct(d_data)
        ref = self.get_ref()
        ref = self.crop_array(ref, config)
        err_max = np.max(np.abs(res - ref))
        assert err_max < self.tol, "something wrong with reconstructor, config = %s" % str(config)


    def get_ref(self):
        if self.ref is not None:
            return self.ref
        if __has_cuda_fbp__:
            fbp_cls = CudaBackprojector
        ref = np.zeros((512, 512, 512), "f")
        fbp = fbp_cls((self.projs.shape[0], self.projs.shape[-1]))
        for i in range(512):
            ref[i] = fbp.fbp(self.d_projs[:, i, :])
        self.ref = ref
        return self.ref


    @staticmethod
    def crop_array(arr, config):
        indices = config["indices"]
        axis = config["axis"]
        slices_roi = config["slices_roi"] or (None, None, None, None)
        i_slice = slice(*indices)
        u_slice = slice(*slices_roi[:2])
        v_slice = slice(*slices_roi[-2:])
        if axis == "z":
            z_slice, y_slice, x_slice = i_slice, v_slice, u_slice
        if axis == "y":
            z_slice, y_slice, x_slice = v_slice, i_slice, u_slice
        if axis == "x":
            z_slice, y_slice, x_slice = v_slice, u_slice, i_slice
        return arr[z_slice, y_slice, x_slice]





import numpy as np
from ...cuda.utils import __has_pycuda__, __pycuda_error_msg__, get_cuda_context
from ...preproc.ccd_cuda import CudaLog, CudaCCDFilter
from ...preproc.flatfield_cuda import CudaFlatFieldDataUrls
from ...preproc.shift_cuda import CudaVerticalShift
from ...preproc.ctf_cuda import CudaCTFPhaseRetrieval
from ...preproc.ctf import GeoPars
from ...preproc.double_flatfield_cuda import CudaDoubleFlatField
from ...preproc.phase_cuda import CudaPaganinPhaseRetrieval
from ...reconstruction.sinogram import SinoBuilder
from ...reconstruction.sinogram_cuda import CudaSinoBuilder, CudaSinoNormalization
from ...reconstruction.rings_cuda import CudaMunchDeringer
from ...misc.unsharp_cuda import CudaUnsharpMask
from ...misc.rotation_cuda import CudaRotation
from ...misc.histogram_cuda import CudaPartialHistogram
from ...reconstruction.fbp import Backprojector
from .grouped import GroupedPipeline, SinoStackPipeline
from .chunked_cuda import CudaChunkedPipeline
from ..utils import pipeline_step

class CudaGroupedPipeline(GroupedPipeline):
    """
    Cuda backend of GroupedPipeline
    """

    backend = "cuda"
    FlatFieldClass = CudaFlatFieldDataUrls
    DoubleFlatFieldClass = CudaDoubleFlatField
    CCDCorrectionClass = CudaCCDFilter
    PaganinPhaseRetrievalClass = CudaPaganinPhaseRetrieval
    UnsharpMaskClass = CudaUnsharpMask
    ImageRotationClass = CudaRotation
    VerticalShiftClass = CudaVerticalShift
    SinoBuilderClass = CudaSinoBuilder
    SinoDeringerClass = CudaMunchDeringer
    MLogClass = CudaLog
    FBPClass = Backprojector
    HistogramClass = CudaPartialHistogram
    SinoNormalizationClass = CudaSinoNormalization

    def __init__(self, process_config, images_group, sub_region=None, logger=None, extra_options=None, cuda_options=None):
        self._init_cuda(cuda_options)
        super().__init__(
             process_config, images_group,
             sub_region=sub_region, logger=logger, extra_options=extra_options
        )

    # TODO base class ?
    def _init_cuda(self, cuda_options):
        if not(__has_pycuda__):
            raise ImportError(__pycuda_error_msg__)
        cuda_options = cuda_options or {}
        self.ctx = get_cuda_context(**cuda_options)
        # self._d_radios = None

    _allocate_array = CudaChunkedPipeline._allocate_array # TODO base class ?


    def _init_pipeline(self):
        super()._init_pipeline()
        self._allocate_array(self.radios.shape, "f", name="radios")


    # overwrite
    def _init_ctf_phase(self):
        options = self.processing_options["phase"]
        geo_pars_params = options["ctf_geo_pars"].copy()
        geo_pars_params["logger"] = self.logger
        geo_pars = GeoPars(**geo_pars_params)
        self.phase_retrieval = CudaCTFPhaseRetrieval(
            self.radios.shape[1:],
            geo_pars,
            options["delta_beta"],
            lim1=options["ctf_lim1"],
            lim2=options["ctf_lim2"],
            logger=self.logger
        )


    @pipeline_step("phase_retrieval", "Performing phase retrieval")
    def _retrieve_phase_ctf(self):
        options = self.processing_options["phase"]
        padding_mode = options["padding_type"]
        for i in range(self.radios.shape[0]):
            self.phase_retrieval.retrieve_phase(
                self.radios[i],
                output=self.radios[i],
                normalize_by_mean=options["ctf_normalize_by_mean"],
            )


    def _read_data(self):
        super()._read_data()
        self._d_radios.set(self.radios)
        self._h_radios = self.radios
        self.radios = self._d_radios

    def _write_data(self, data=None):
        self._d_radios.get(ary=self._h_radios)
        self.radios = self._h_radios
        super()._write_data(data=self.radios)



class CudaSinoStackPipeline(SinoStackPipeline):
    backend = "cuda"
    SinoBuilderClass = SinoBuilder # done on host to save device memory
    SinoNormalizationClass = CudaSinoNormalization
    SinoDeringerClass = CudaMunchDeringer
    FBPClass = Backprojector
    HistogramClass = CudaPartialHistogram

    def __init__(self, process_config, stack, projections, logger=None, extra_options=None, cuda_options=None):
        self._init_cuda(cuda_options)
        super().__init__(
             process_config, stack, projections, logger=logger, extra_options=extra_options,
        )

    # TODO base class ?
    _init_cuda = CudaGroupedPipeline._init_cuda
    _allocate_array = CudaChunkedPipeline._allocate_array
    _allocate_array_on_host = GroupedPipeline._allocate_array

    def _allocate_sinobuilder_output(self):
        # build_sino is done on host. Use host allocation
        return self._allocate_array_on_host(self.sino_builder.output_shape, "f", name="sinos")

    def _build_sino(self):
        super()._build_sino()
        self._allocate_array(self.sino_builder.output_shape, "f", name="sinos")
        self._d_sinos.set(np.ascontiguousarray(self.sinos)) # !
        self._h_sinos = self.sinos
        self.sinos = self._d_sinos


    def _write_data(self, data=None):
        recs = self._d_recs.get() # not ideal - use self.recs ?
        super()._write_data(data=recs)



import numpy as np
from ..thirdparty.pore3d_deringer_munch import munchetal_filter
from ..utils import deprecation_warning
from .sinogram import SinoBuilder

class MunchDeringer:
    def __init__(self, sigma, sinos_shape=None, radios_shape=None, levels=None, wname='db15'):
        """
        Initialize a "Munch Et Al" sinogram deringer. See References for more information.

        Parameters
        -----------
        sigma: float
            Standard deviation of the damping parameter. The higher value of sigma,
            the more important the filtering effect on the rings.
        levels: int, optional
            Number of wavelets decomposition levels.
            By default (None), the maximum number of decomposition levels is used.
        wname: str, optional
            Default is "db15" (Daubechies, 15 vanishing moments)
        sinos_shape: tuple, optional
            Shape of the sinogram (or sinograms stack).
            This class requires either sinos_shape or radios_shape.
        radios_shape: tuple, optional
            Shape of the projection (or projections stack)
            This class requires either sinos_shape or radios_shape.

        References
        ----------
        B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
        combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
        """
        if radios_shape is not None:
            deprecation_warning(
                "The 'radios_shape' parameter is now ignored",
                do_print=True, func_name="munchderinger_init"
            )
        self._get_shapes(sinos_shape=sinos_shape, radios_shape=radios_shape)
        self.sigma = sigma
        self.levels = levels
        self.wname = wname
        self._check_can_use_wavelets()


    _get_shapes = SinoBuilder._get_shapes

    def _check_can_use_wavelets(self):
        if munchetal_filter is None:
            raise ValueError("Need pywavelets to use this class")


    def _destripe_2D(self, sino, output):
        res = munchetal_filter(sino, self.levels, self.sigma, wname=self.wname)
        output[:] = res
        return output


    def remove_rings(self, sinos, output=None):
        """
        Main function to performs rings artefacts removal on sinogram(s).
        CAUTION: this function defaults to in-place processing, meaning that
        the sinogram(s) you pass will be overwritten.

        Parameters
        ----------
        sinos: numpy.ndarray
            Sinogram or stack of sinograms.
        output: numpy.ndarray, optional
            Output array. If set to None (default), the output overwrites the input.
        """
        if output is None:
            output = sinos
        if sinos.ndim == 2:
            return self._destripe_2D(sinos, output)
        n_sinos = sinos.shape[0]
        for i in range(n_sinos):
            self._destripe_2D(sinos[i], output[i])
        return output



import posixpath
from os import path
from math import ceil
from shutil import copy
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
import numpy as np
from tomoscan.io import HDF5File
from tomoscan.esrf.hdf5scan import HDF5TomoScan
from ..io.utils import get_first_hdf5_entry
from ..misc.rotation import Rotation
from ..resources.logger import Logger, LoggerOrPrint
from ..pipeline.config_validators import optional_tuple_of_floats_validator, boolean_validator
from ..misc.rotation_cuda import CudaRotation, __has_pycuda__
from .utils import parse_params_values
from .cli_configs import RotateRadiosConfig


class HDF5ImagesStackRotation:
    def __init__(self, input_file, output_file, angle, center=None, entry=None, logger=None, batch_size=100, use_cuda=True, use_multiprocessing=True):
        self.logger = LoggerOrPrint(logger)
        self.use_cuda = use_cuda & __has_pycuda__
        self.batch_size = batch_size
        self.use_multiprocessing = use_multiprocessing
        self._browse_dataset(input_file, entry)
        self._get_rotation(angle, center)
        self._init_output_dataset(output_file)

    def _browse_dataset(self, input_file, entry):
        self.input_file = input_file
        if entry is None or entry == "":
            entry = get_first_hdf5_entry(input_file)
        self.entry = entry
        self.dataset_info = HDF5TomoScan(input_file, entry=entry)

    def _get_rotation(self, angle, center):
        if self.use_cuda:
            self.logger.info("Using Cuda rotation")
            rot_cls = CudaRotation
        else:
            self.logger.info("Using skimage rotation")
            rot_cls = Rotation
            if self.use_multiprocessing:
                self.thread_pool = ThreadPool(processes=cpu_count() - 2)
                self.logger.info("Using multiprocessing with %d cores" % self.thread_pool._processes)

        self.rotation = rot_cls(
            (self.dataset_info.dim_2, self.dataset_info.dim_1),
            angle,
            center=center,
            mode="edge"
        )

    def _init_output_dataset(self, output_file):
        self.output_file = output_file
        copy(self.input_file, output_file)

        first_proj_url = self.dataset_info.projections[list(self.dataset_info.projections.keys())[0]]
        self.data_path = first_proj_url.data_path()
        dirname, basename = posixpath.split(self.data_path)
        self._data_path_dirname = dirname
        self._data_path_basename = basename

    def _rotate_stack_cuda(self, images, output):
        self.rotation.cuda_processing._allocate_array("tmp_images_stack", images.shape)
        self.rotation.cuda_processing._allocate_array("tmp_images_stack_rot", images.shape)
        d_in = self.rotation.cuda_processing.tmp_images_stack
        d_out = self.rotation.cuda_processing.tmp_images_stack_rot
        n_imgs = images.shape[0]
        d_in[:n_imgs].set(images)
        for j in range(n_imgs):
            self.rotation.rotate(d_in[j], output=d_out[j])
        d_out[:n_imgs].get(ary=output[:n_imgs])

    def _rotate_stack(self, images, output):
        if self.use_cuda:
            self._rotate_stack_cuda(images, output)
        elif self.use_multiprocessing:
            out_tmp = self.thread_pool.map(self.rotation.rotate, images)
            print(out_tmp[0])
            output[:] = np.array(out_tmp, dtype="f") # list -> np array... consumes twice as much memory
        else:
            for j in range(images.shape[0]):
                output[j] = self.rotation.rotate(images[j])

    def rotate_images(self, suffix="_rot"):
        data_path = self.data_path
        fid = HDF5File(self.input_file, "r")
        fid_out = HDF5File(self.output_file, "a")

        try:
            data_ptr = fid[data_path]
            n_images = data_ptr.shape[0]
            data_out_ptr = fid_out[data_path]

            # Delete virtual dataset in output file, create "data_rot" dataset
            del fid_out[data_path]
            fid_out[self._data_path_dirname].create_dataset(
                self._data_path_basename + suffix, shape=data_ptr.shape, dtype=data_ptr.dtype
            )
            data_out_ptr = fid_out[data_path + suffix]

            # read by group of images to hide latency
            group_size = self.batch_size
            images_rot = np.zeros((group_size, data_ptr.shape[1], data_ptr.shape[2]), dtype="f")
            n_groups = ceil(n_images / group_size)
            for i in range(n_groups):
                self.logger.info("Processing radios group %d/%d" % (i+1, n_groups))
                i_min = i * group_size
                i_max = min((i + 1) * group_size, n_images)
                images = data_ptr[i_min:i_max, :, :].astype("f")
                self._rotate_stack(images, images_rot)
                data_out_ptr[i_min:i_max, :, :] = images_rot[:i_max-i_min, :, :].astype(data_ptr.dtype)
        finally:
            fid_out[self._data_path_dirname].move(posixpath.basename(data_path) + suffix, self._data_path_basename)
            fid_out[data_path].attrs["interpretation"] = "image"
            fid.close()
            fid_out.close()




def rotate_cli():
    args = parse_params_values(
        RotateRadiosConfig,
        parser_description="A command-line utility for performing a rotation on all the radios of a dataset."
    )
    logger = Logger(
        "nabu_rotate", level=args["loglevel"], logfile="nabu_rotate.log"
    )

    dataset_path = args["dataset"]
    h5_entry = args["entry"]
    output_file = args["output"]
    center = optional_tuple_of_floats_validator("", "", args["center"])
    use_cuda = boolean_validator("", "", args["use_cuda"])
    use_multiprocessing = boolean_validator("", "", args["use_multiprocessing"])

    if path.exists(output_file):
        logger.fatal("Output file %s already exists, not overwriting it" % output_file)
        exit(1)

    h5rot = HDF5ImagesStackRotation(
        dataset_path, output_file, args["angle"], center=center, entry=h5_entry, logger=logger,
        batch_size=args["batchsize"], use_cuda=use_cuda, use_multiprocessing=use_multiprocessing
    )
    h5rot.rotate_images()


if __name__ == "__main__":
    rotate_cli()



#!/usr/bin/env python

"""
computations.py: determine computational needs, chunking method to be used, etc.
"""

from silx.image.tomography import get_next_power


def estimate_required_memory(process_config, chunk_size=None, group_size=None, radios_and_sinos=True, warn_from_GB=None):
    """
    Estimate the memory (RAM) needed for a reconstruction.

    Parameters
    -----------
    process_config: `ProcessConfig` object
        Data structure with the processing configuration
    chunk_size: int, optional
        Size of a "radios chunk", i.e "delta z". A radios chunk is a 3D array of shape (n_angles, chunk_size, n_x)
        If set to None, then chunk_size = n_z
    group_size: int, optional
        Size of a "radios group", i.e "num_angles". A radios group is a 3D array of shape (group_size, n_z, n_x)
        If set to None, then group_size = n_angles
    radios_and_sinos: bool
        Whether radios and sinograms will co-exist in memory (meaning more memory usage)
    warn_from_GB: float, optional
        Amount of memory in GB from where a warning flag will be raised.
        Default is None
        If set to a number, the result will be in the form (estimated_memory_GB, warning)
        where 'warning' is a boolean indicating whether memory allocation/transfer might be problematic.

    Notes
    -----
    It seems that Cuda does not allow allocating and/or transferring more than 16384 MiB (17.18 GB).
    If warn_from_GB is not None, then the result is in the form (estimated_memory_GB, warning)
    where warning is a boolean indicating wheher memory allocation/transfer might be problematic.
    """
    dataset = process_config.dataset_info
    nabu_config = process_config.nabu_config
    processing_steps = process_config.processing_steps
    Nx, Ny = dataset.radio_dims
    if chunk_size is not None:
        Ny = chunk_size
    Na = dataset.n_angles
    if group_size is not None:
        Na = group_size

    total_memory_needed = 0
    memory_warning = False

    # Read data
    # ----------
    binning_z = nabu_config["dataset"]["binning_z"]
    projections_subsampling = nabu_config["dataset"]["projections_subsampling"]
    data_volume_size = Nx * Ny * Na * 4
    data_image_size = Nx * Ny * 4
    total_memory_needed += data_volume_size
    if (warn_from_GB is not None) and data_volume_size/1e9 > warn_from_GB:
        memory_warning = True

    # CCD processing
    # ---------------
    if "flatfield" in processing_steps:
        # Flat-field is done in-place, but still need to load darks/flats
        n_darks = len(dataset.darks)
        n_flats = len(dataset.flats)
        darks_size = n_darks * Nx * Ny * 2  # uint16
        flats_size = n_flats * Nx * Ny * 4  # f32
        total_memory_needed += darks_size + flats_size

    if "ccd_correction" in processing_steps:
        # CCD filter is "batched 2D"
        total_memory_needed += data_image_size

    # Phase retrieval
    # ---------------
    if "phase" in processing_steps:
        # Phase retrieval is done image-wise, so near in-place, but needs to
        # allocate some images, fft plans, and so on
        Nx_p = get_next_power(2 * Nx)
        Ny_p = get_next_power(2 * Ny)
        img_size_real = 2 * 4 * Nx_p * Ny_p
        img_size_cplx = 2 * 8 * ((Nx_p * Ny_p) // 2 + 1)
        total_memory_needed += 2 * img_size_real + 3 * img_size_cplx

    # Sinogram de-ringing
    # -------------------
    if "sino_rings_correction" in processing_steps:
        # Process is done image-wise.
        # Needs one Discrete Wavelets transform and one FFT/IFFT plan for each scale
        total_memory_needed += (Nx * Na * 4) * 5.5  # approx.


    # Reconstruction
    # ---------------
    reconstructed_volume_size = 0
    if "reconstruction" in processing_steps:
        if radios_and_sinos:
            total_memory_needed += data_volume_size  # radios + sinos
        rec_config = process_config.processing_options["reconstruction"]
        if "rotation_axis_position_halftomo" in rec_config:
            # Slice has a different shape in half acquisition.
            rc = rec_config["rotation_axis_position_halftomo"]
            Nx_rec = 2 * rc
            Nx_rec = max(2 * rc, 2 * (Nx - rc))
            Ny_rec = Nx_rec
        else:
            Nx_rec = (rec_config["end_x"] - rec_config["start_x"] + 1)
            Ny_rec = (rec_config["end_y"] - rec_config["start_y"] + 1)
        Nz_rec = (rec_config["end_z"] - rec_config["start_z"] + 1) // binning_z
        if chunk_size:
            Nz_rec = chunk_size // binning_z
        reconstructed_volume_size = Nx_rec * Ny_rec * Nz_rec * 4  # float32
        total_memory_needed += reconstructed_volume_size

    if warn_from_GB is None:
        return total_memory_needed
    else:
        return (total_memory_needed, memory_warning)



def estimate_chunk_size(available_memory_GB, process_config, chunk_step=50, warn_from_GB=None):
    """
    Estimate the maximum chunk size given an avaiable amount of memory.

    Parameters
    -----------
    available_memory_GB: float
        available memory in Giga Bytes (GB - not GiB !).
    process_config: ProcessConfig
        ProcessConfig object
    """
    chunk_size = chunk_step
    radios_and_sinos = False
    if (
        "reconstruction" in process_config.processing_steps
        and process_config.processing_options["reconstruction"]["enable_halftomo"]
    ):
        radios_and_sinos = True

    max_dz = process_config.dataset_info.radio_dims[1]
    chunk_size = chunk_step
    last_good_chunk_size = chunk_size
    while True:
        res = estimate_required_memory(
            process_config,
            chunk_size=chunk_size,
            radios_and_sinos=radios_and_sinos,
            warn_from_GB=warn_from_GB # 2**32 elements - see estimate_required_memory docstring note
        )
        if warn_from_GB is not None:
            (required_mem, mem_warn) = res
        else:
            (required_mem, mem_warn) = res, False
        required_mem_GB = required_mem / 1e9
        if required_mem_GB > available_memory_GB or chunk_size > max_dz or mem_warn:
            break
        last_good_chunk_size = chunk_size
        chunk_size += chunk_step
    return last_good_chunk_size


def estimate_group_size(available_memory_GB, process_config, step=50, warn_from_GB=None):
    """
    Same as estimate_chunk_size, but for radios group
    """

    def _remove_sino_processing_steps(processing_steps):
        sino_steps = ["sino_rings_correction", "reconstruction"]
        removed_sino_steps = {}
        for step in sino_steps:
            if step in processing_steps:
                idx = processing_steps.index(step)
                removed_sino_steps[idx] = processing_steps.pop(idx)
        return removed_sino_steps

    def _restore_sino_processing_steps(processing_steps, removed_sino_steps):
        for idx, val in removed_sino_steps.items():
            processing_steps.insert(idx, val)

    removed_sino_steps = _remove_sino_processing_steps(process_config.processing_steps)

    group_size = step
    n_angles = process_config.dataset_info.n_angles
    last_good_group_size = group_size
    while True:
        res = estimate_required_memory(
            process_config,
            group_size=group_size,
            radios_and_sinos=False,
            warn_from_GB=warn_from_GB # 2**32 elements - see estimate_required_memory docstring note
        )
        if warn_from_GB is not None:
            (required_mem, mem_warn) = res
        else:
            (required_mem, mem_warn) = res, False
        required_mem_GB = required_mem / 1e9
        if required_mem_GB > available_memory_GB or group_size > n_angles or mem_warn:
            break
        last_good_group_size = group_size
        group_size += step
    _restore_sino_processing_steps(process_config.processing_steps, removed_sino_steps)
    return last_good_group_size




import inspect
import numpy as np
from silx.io import get_data
from ..preproc.flatfield import FlatFieldDataUrls
from ..estimation.cor import CenterOfRotation, CenterOfRotationAdaptiveSearch, CenterOfRotationSlidingWindow, CenterOfRotationGrowingWindow
from ..estimation.cor_sino import SinoCor
from ..estimation.tilt import CameraTilt
from ..estimation.utils import is_fullturn_scan
from ..resources.logger import LoggerOrPrint
from ..resources.utils import extract_parameters
from ..utils import check_supported, is_int, deprecation_warning
from .params import tilt_methods
from ..resources.dataset_analyzer import get_0_180_radios
from ..misc.rotation import Rotation
from ..io.reader import ChunkReader
from ..preproc.ccd import Log

"""
nabu.pipeline.estimators: helper classes/functions to estimate parameters of a dataset
(center of rotation, detector tilt, etc).
"""

class CORFinder:
    """
    An application-type class for finding the Center Of Rotation (COR).
    """

    search_methods = {
        "centered":{
            "class": CenterOfRotation,
        },
        "global": {
            "class": CenterOfRotationAdaptiveSearch,
            "default_kwargs": {"low_pass": 1, "high_pass":20},
        },
        "sliding-window": {
            "class": CenterOfRotationSlidingWindow,
            "default_args": ["center"],
        },
        "growing-window": {
            "class": CenterOfRotationGrowingWindow,
        },
    }

    def __init__(self, dataset_info, angles=None, halftomo=False, do_flatfield=True, cor_options=None, logger=None):
        """
        Initialize a CORFinder object.

        Parameters
        ----------
        dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
            Dataset information structure
        """
        # COMPAT.
        deprecation_warning(
            "The parameters 'angles' and 'halftomo' are now ignored and deprecated.",
            func_name="corfinder"
        )
        # ---
        self.logger = LoggerOrPrint(logger)
        self.dataset_info = dataset_info
        self.do_flatfield = do_flatfield
        self.shape = dataset_info._radio_dims_notbinned[::-1]
        self._init_radios()
        self._init_flatfield()
        self._apply_flatfield()
        self._apply_tilt()
        self._default_search_method = "centered"
        if self.dataset_info.is_halftomo:
            self._default_search_method = "sliding-window"
        self._get_cor_options(cor_options)


    def _init_radios(self):
        self.radios, self._radios_indices = get_0_180_radios(self.dataset_info, return_indices=True)


    def _init_flatfield(self):
        if not(self.do_flatfield):
            return
        self.flatfield = FlatFieldDataUrls(
            self.radios.shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self._radios_indices,
            interpolation="linear",
            convert_float=True
        )


    def _apply_flatfield(self):
        if not(self.do_flatfield):
            return
        self.flatfield.normalize_radios(self.radios)


    def _apply_tilt(self):
        tilt = self.dataset_info.detector_tilt
        if tilt is None:
            return
        self.logger.debug("COREstimator: applying detector tilt correction of %f degrees" % tilt)
        rot = Rotation(self.shape, tilt)
        for i in range(self.radios.shape[0]):
            self.radios[i] = rot.rotate(self.radios[i])


    def _get_cor_options(self, cor_options):
        if cor_options is None:
            self.cor_options = {}
            return
        try:
            cor_options = extract_parameters(cor_options)
        except Exception as exc:
            msg = "Could not extract parameters from cor_options: %s" % (str(exc))
            self.logger.fatal(msg)
            raise ValueError(msg)
        self.cor_options = cor_options


    def find_cor(self, method=None):
        """
        Find the center of rotation.

        Parameters
        ----------
        method: str, optional
            Which CoR search method to use. Default "centered".

        Returns
        -------
        cor: float
            The estimated center of rotation for the current dataset.

        Notes
        ------
        This function passes the named parameters to nabu.preproc.alignment.CenterOfRotation.find_shift.
        """
        method = method or self._default_search_method
        check_supported(method, self.search_methods.keys(), "CoR estimation method")
        cor_class = self.search_methods[method]["class"]
        cor_finder = cor_class(logger=self.logger)
        self.logger.info("Estimating center of rotation")

        default_params = self.search_methods[method].get("default_kwargs", None) or {}
        cor_exec_kwargs = default_params.copy()
        cor_exec_kwargs.update(self.cor_options)
        cor_exec_args = self.search_methods[method].get("default_args", None) or []
        # Specific to CenterOfRotationSlidingWindow
        if cor_class == CenterOfRotationSlidingWindow:
            side_param = cor_exec_kwargs.pop("side", "center")
            cor_exec_args = [side_param]
        #
        cor_exec_kwargs.pop("slice", None)
        self.logger.debug("%s(%s)" % (get_class_name(cor_class), str(cor_exec_kwargs)))
        shift = cor_finder.find_shift(
            self.radios[0],
            np.fliplr(self.radios[1]),
            *cor_exec_args,
            **cor_exec_kwargs
        )
        # find_shift returned a single scalar in 2020.1
        # This should be the default after 2020.2 release
        if hasattr(shift, "__iter__"):
            shift = shift[0]
        #
        res = self.shape[1]/2 + shift
        self.logger.info("Estimated center of rotation: %.2f" % res)
        return res

# alias
COREstimator = CORFinder

class SinoCORFinder:
    """
    A class for finding Center of Rotation based on 360 degrees sinograms.
    This class handles the steps of building the sinogram from raw radios.
    """

    search_methods = ["sino-coarse-to-fine", "sliding-window", "growing-window"]
    default_method = "sino-coarse-to-fine"

    def __init__(self, dataset_info, slice_idx, subsampling=10, do_flatfield=True, cor_options=None, logger=None):
        """
        Initialize a SinoCORFinder object.

        Parameters
        ---------
        dataset_info: `nabu.resources.dataset_analyzer.DatasetAnalyzer`
            Dataset information structure
        slice_idx: int or str
            Which slice index to take for building the sinogram.
            For example slice_idx=0 means that we extract the first line of each projection.
            Value can also be "first", "top", "middle", "last", "bottom".
        subsampling: int, float
            subsampling strategy when building sinograms.
            As building the complete sinogram from raw projections might be tedious, the reading is done with subsampling.
            A positive integer value means the subsampling step (i.e `projections[::subsampling]`).
            A negative integer value means we take -subsampling projections in total.
            A float value indicates the angular step in DEGREES.
        do_flatfield: bool, optional
            Whether to perform flat-field normalization. Default is True.
        cor_options: str
            User options for the auto-CoR method.
        logger: Logger, optional
            Logging object
        """
        self.logger = LoggerOrPrint(logger)
        self.dataset_info = dataset_info
        self._check_360()
        self._set_slice_idx(slice_idx)
        self._set_subsampling(subsampling)
        self._load_raw_sinogram()
        self._flatfield(do_flatfield)
        self._get_sinogram()
        self._get_cor_options(cor_options)

    _get_cor_options = CORFinder._get_cor_options


    def _check_360(self):
        if self.dataset_info.dataset_scanner.scan_range == 360:
            return
        if not is_fullturn_scan(self.dataset_info.rotation_angles):
            raise ValueError("Sinogram-based Center of Rotation estimation can only be used for 360 degrees scans")


    def _set_slice_idx(self, slice_idx):
        n_z = self.dataset_info.radio_dims[1]
        if isinstance(slice_idx, str):
            str_to_idx = {
                "top": 0,
                "first": 0,
                "middle": n_z // 2,
                "bottom": n_z - 1,
                "last": n_z - 1
            }
            check_supported(slice_idx, str_to_idx.keys(), "slice location")
            slice_idx = str_to_idx[slice_idx]
        self.slice_idx = slice_idx


    def _set_subsampling(self, subsampling):
        projs_idx = sorted(self.dataset_info.projections.keys())
        if is_int(subsampling):
            if subsampling < 0: # Total number of angles
                n_angles = -subsampling
                indices_float = np.linspace(projs_idx[0], projs_idx[-1], n_angles, endpoint=True)
                self.projs_indices = np.round(indices_float).astype(np.int32).tolist()
            else: # Subsampling step
                self.projs_indices = projs_idx[::subsampling]
        else: # Angular step
            raise NotImplementedError()


    def _load_raw_sinogram(self):
        # Subsample projections
        files = {}
        for idx in self.projs_indices:
            files[idx] = self.dataset_info.projections[idx]
        self.files = files
        self.data_reader = ChunkReader(
            self.files,
            sub_region=(None, None, self.slice_idx, self.slice_idx+1),
            convert_float=True,
        )
        self.data_reader.load_files()
        self._radios = self.data_reader.files_data


    def _flatfield(self, do_flatfield):
        self.do_flatfield = bool(do_flatfield)
        if not self.do_flatfield:
            return
        flatfield = FlatFieldDataUrls(
            self._radios.shape,
            self.dataset_info.flats,
            self.dataset_info.darks,
            radios_indices=self.projs_indices,
            sub_region=(None, None, self.slice_idx, self.slice_idx+1)
        )
        flatfield.normalize_radios(self._radios)


    def _get_sinogram(self):
        log = Log(self._radios.shape, clip_min=1e-6, clip_max=10.)
        sinogram = self._radios[:, 0, :].copy()
        log.take_logarithm(sinogram)
        self.sinogram = sinogram


    @staticmethod
    def _split_sinogram(sinogram):
        n_a_2 = sinogram.shape[0]//2
        img_1, img_2 = sinogram[:n_a_2], sinogram[n_a_2:]
        return img_1, img_2


    def _find_cor_sliding_window(self):
        cor_finder = CenterOfRotationSlidingWindow(logger=self.logger)

        img_1, img_2 = self._split_sinogram(self.sinogram)
        kwargs = update_func_kwargs(cor_finder.find_shift, self.cor_options)
        side = self.cor_options.get("side", "right")
        kwargs.pop("side", None)
        self.logger.debug("CenterOfRotationSlidingWindow.find_shift(%s)" % str(kwargs))
        cor = cor_finder.find_shift(img_1, img_2, side, **kwargs)
        return cor[0] + self.sinogram.shape[1] / 2.


    def _find_cor_growing_window(self):
        cor_finder = CenterOfRotationGrowingWindow(logger=self.logger)

        img_1, img_2 = self._split_sinogram(self.sinogram)
        kwargs = update_func_kwargs(cor_finder.find_shift, self.cor_options)
        self.logger.debug("CenterOfRotationGrowingWindow.find_shift(%s)" % str(kwargs))
        cor = cor_finder.find_shift(img_1, img_2, **kwargs)

        return cor[0] + self.sinogram.shape[1] / 2.


    def _find_cor_coarse2fine(self):
        side = self.cor_options.get("side", "right")
        window_width = self.cor_options.get("window_width", None)
        neighborhood = self.cor_options.get("neighborhood", 7)
        shift_value = self.cor_options.get("shift_value", 0.1)
        cor_finder = SinoCor(self.sinogram, logger=self.logger)
        self.logger.debug(
            "SinoCor.estimate_cor_coarse(side=%s, window_width=%s)"
            % (str(side), str(window_width))
        )
        cor_finder.estimate_cor_coarse(side=side, window_width=window_width)
        self.logger.debug(
            "SinoCor.estimate_cor_fine(neighborhood=%s, shift_value=%s)"
            % (str(neighborhood), str(shift_value))
        )
        cor = cor_finder.estimate_cor_fine(neighborhood=neighborhood, shift_value=shift_value)
        return cor


    def find_cor(self, method=None):
        method = method or self.default_method
        cor_estimation_function = {
            "sino-coarse-to-fine": self._find_cor_coarse2fine,
            "sliding-window": self._find_cor_sliding_window,
            "growing-window": self._find_cor_growing_window,
        }
        check_supported(method, cor_estimation_function.keys(), "sinogram-based CoR estimation method")
        res = cor_estimation_function[method]()
        return res

# alias
SinoCOREstimator = SinoCORFinder


class CompositeCORFinder:
    """
    Class and method to prepare sinogram and calculate COR in HA
    The pseudo sinogram is built with shrinked radios taken every theta degres
    """
    def __init__(self, dataset_info, oversampling=4, theta=10, subsampling_y=10, take_log=True, cor_options=None, logger=None):

        self.dataset_info = dataset_info
        self.logger = LoggerOrPrint(logger)
        self._check_360()
        self._get_cor_options(cor_options)

        self.take_log = take_log
        self.ovs = oversampling
        self.theta = theta
        self.yred_fact = subsampling_y

        self.nproj = self.dataset_info.n_angles

        self.dproj = round(self.nproj/360*self.theta)
        self.proj_stop = round(self.nproj/(2*self.dproj)) #+ 1
        self.sx, self.sy = self.dataset_info.radio_dims
        self.rcor_abs = round(self.sx/2.)
        self.cor_acc = round(self.sx/2.)

        # in this mode, the sinogram will be composed by a succession of 360/dtheta radios
        # 180/dtheta should be integer. sy/syred_fact should be integer
        # The radios themseleves are compressed verticaly by a factor of yred_fact.
        # in the original Paul's algorithm, the radio are horizontally oversampled by 4 to obtain
        # the subpixel precision. Here we prefor to do a float schift on 1/10 pixels to obtain this
        # subprecision.

        self.nprojred = 2*self.proj_stop
        self.yred = round(self.sy/self.yred_fact)

        # initialize sinograms and radios arrays
        self.sino = np.zeros((int(self.nprojred * self.yred), self.sx))
        self._loaded = False

        self.projs_indices = np.arange(self.proj_stop)*self.dproj
        self.projs_absolute_indices = sorted(self.dataset_info.projections.keys())

        self.flatfield = FlatFieldDataUrls(
            (len(self.projs_indices), self.sy, self.sx),
            self.dataset_info.flats,
            self.dataset_info.darks,
            radios_indices=[self.projs_absolute_indices[i] for i in self.projs_indices],
            dtype=np.float64
        )
        self.mlog = Log(
            (1, ) + self.flatfield.shape,
            clip_min=1e-6,
            clip_max=10.
        )


    def get_radio(self, image_num):
        radio_dataset_idx = self.projs_absolute_indices[image_num]
        data_url = self.dataset_info.projections[radio_dataset_idx]
        radio = get_data(data_url).astype(np.float64)
        self.flatfield.normalize_single_radio(radio, radio_dataset_idx, dtype=radio.dtype)
        if self.take_log:
            self.mlog.take_logarithm(radio)
        return radio


    def get_sino(self, reload=False):
        """
        Build sinogram (composite image) from the radio files
        """
        if self._loaded and not reload:
            return self.sino
        np2 = round(self.nproj/2)
        ns2 = round(self.sino.shape[0]/2)
        # loop on all the projections
        irad = 0
        for npi in self.projs_indices:
            radio1 = np.resize(self.get_radio(npi), (self.yred, self.sx))
            radio2 = np.resize(self.get_radio(npi+np2), (self.yred, self.sx))

            self.sino[irad:irad+self.yred,:] = radio1
            self.sino[irad+ns2:irad+ns2+self.yred,:] = radio2

            irad = irad + self.yred

        self.sino[np.isnan(self.sino)] = 0.0001 # ?
        return self.sino

    _check_360 = SinoCORFinder._check_360
    _get_cor_options = SinoCORFinder._get_cor_options
    _find_cor_coarse2fine = SinoCORFinder._find_cor_coarse2fine

    def find_cor(self, reload=False, **kwargs):
        self.sinogram = self.get_sino(reload=reload)
        return self._find_cor_coarse2fine()


# alias
CompositeCOREstimator = CompositeCORFinder


def get_function_default_kwargs(func):
    defaults_kwargs_vals = func.__defaults__
    varnames = func.__code__.co_varnames

# Some heavily inelegant things going on here
def get_default_kwargs(func):
    params = inspect.signature(func).parameters
    res = {}
    for param_name, param in params.items():
        if param.default != inspect._empty:
            res[param_name] = param.default
    return res


def update_func_kwargs(func, options):
    res_options = get_default_kwargs(func)
    for option_name, option_val in options.items():
        if option_name in res_options:
            res_options[option_name] = options[option_name]
    return res_options


def get_class_name(class_object):
    return str(class_object).split(".")[-1].strip(">").strip("'").strip('"')



class DetectorTiltEstimator:
    """
    Helper class for detector tilt estimation.
    It automatically chooses the right radios and performs flat-field.
    """
    default_tilt_method = "1d-correlation"
    # Given a tilt angle "a", the maximum deviation caused by the tilt (in pixels) is
    #  N/2 * |sin(a)|  where N is the number of pixels
    # We ignore tilts causing less than 0.25 pixel deviation: N/2*|sin(a)| < tilt_threshold
    tilt_threshold = 0.25

    def __init__(self, dataset_info, do_flatfield=True, logger=None, autotilt_options=None):
        """
        Initialize a detector tilt estimator helper.

        Parameters
        ----------
        dataset_info: `dataset_info` object
            Data structure with the dataset information.
        do_flatfield: bool, optional
            Whether to perform flat field on radios.
        logger: `Logger` object, optional
            Logger object
        autotilt_options: dict, optional
            named arguments to pass to the detector tilt estimator class.
        """
        self._set_params(dataset_info, do_flatfield, logger, autotilt_options)
        self.radios, self.radios_indices = get_0_180_radios(
            dataset_info, return_indices=True
        )
        self._init_flatfield()
        self._apply_flatfield()


    def _set_params(self, dataset_info, do_flatfield, logger, autotilt_options):
        self.dataset_info = dataset_info
        self.do_flatfield = bool(do_flatfield)
        self.logger = LoggerOrPrint(logger)
        self._get_autotilt_options(autotilt_options)


    def _init_flatfield(self):
        if not(self.do_flatfield):
            return
        self.flatfield = FlatFieldDataUrls(
            self.radios.shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self.radios_indices,
            interpolation="linear",
            convert_float=True
        )


    def _apply_flatfield(self):
        if not(self.do_flatfield):
            return
        self.flatfield.normalize_radios(self.radios)


    def _get_autotilt_options(self, autotilt_options):
        if autotilt_options is None:
            self.autotilt_options = None
            return
        try:
            autotilt_options = extract_parameters(autotilt_options)
        except Exception as exc:
            msg = "Could not extract parameters from autotilt_options: %s" % (str(exc))
            self.logger.fatal(msg)
            raise ValueError(msg)
        self.autotilt_options = autotilt_options
        if "threshold" in autotilt_options:
            self.tilt_threshold = autotilt_options.pop("threshold")


    def find_tilt(self, tilt_method=None):
        """
        Find the detector tilt.

        Parameters
        ----------
        tilt_method: str, optional
            Which tilt estimation method to use.
        """
        if tilt_method is None:
            tilt_method = self.default_tilt_method
        check_supported(tilt_method, set(tilt_methods.values()), "tilt estimation method")
        self.logger.info("Estimating detector tilt angle")
        autotilt_params = {
            "roi_yxhw": None,
            "median_filt_shape": None,
            "padding_mode": None,
            "peak_fit_radius": 1,
            "high_pass": None,
            "low_pass": None,
        }
        autotilt_params.update(self.autotilt_options or {})
        self.logger.debug("%s(%s)" % ("CameraTilt", str(autotilt_params)))

        tilt_calc = CameraTilt()
        tilt_cor_position, camera_tilt = tilt_calc.compute_angle(
            self.radios[0],
            np.fliplr(self.radios[1]),
            method=tilt_method,
            **autotilt_params
        )
        self.logger.info("Estimated detector tilt angle: %f degrees" % camera_tilt)
        # Ignore too small tilts
        max_deviation = np.max(self.dataset_info.radio_dims) * np.abs(np.sin(np.deg2rad(camera_tilt)))
        if self.dataset_info.is_halftomo:
            max_deviation *= 2
        if max_deviation < self.tilt_threshold:
            self.logger.info(
                "Estimated tilt angle (%.3f degrees) results in %.2f maximum pixels shift, which is below threshold (%.2f pixel). Ignoring the tilt, no correction will be done."
                % (camera_tilt, max_deviation, self.tilt_threshold)
            )
            camera_tilt = None
        return camera_tilt


# alias
TiltFinder = DetectorTiltEstimator



import numpy as np
from scipy.interpolate import interp1d
from ..utils import get_2D_3D_shape, check_supported, deprecated_class

class SinoBuilder:
    """
    A base class for processing sinograms.
    """

    def __init__(self, sinos_shape=None, radios_shape=None, rot_center=None, halftomo=False, angles=None, interpolate=False):
        """
        Initialize a SinoBuilder instance.

        Parameters
        ----------
        sinos_shape: tuple of int
            Shape of the stack of sinograms, in the form `(n_z, n_angles, n_x)`.
            If not provided, it is derived from `radios_shape`.
        radios_shape: tuple of int
            Shape of the chunk of radios, in the form `(n_angles, n_z, n_x)`.
            If not provided, it is derived from `sinos_shape`.
        rot_center: int or array
            Rotation axis position. A scalar indicates the same rotation axis position
            for all the projections.
        halftomo: bool
            Whether "half tomography" is enabled. Default is False.
        interpolate: bool, optional
            Only used if halftomo=True.
            Whether to re-grid the second part of sinograms to match projection k with projection k + n_a/2.
            This forces each pair of projection (k, k + n_a/2) to be separated by exactly 180 degrees.
        angles: array, optional
            Rotation angles (in radians). Used and required only when halftomo and interpolate are True.
        """
        self._get_shapes(sinos_shape, radios_shape)
        self.set_rot_center(rot_center)
        self._configure_halftomo(halftomo, interpolate, angles)


    def _get_shapes(self, sinos_shape, radios_shape):
        if (sinos_shape is None) and (radios_shape is None):
            raise ValueError("Need to provide sinos_shape and/or radios_shape")
        if sinos_shape is None:
            n_a, n_z, n_x = get_2D_3D_shape(radios_shape)
            sinos_shape = (n_z, n_a, n_x)
        elif len(sinos_shape) == 2:
            sinos_shape = (1, ) + sinos_shape
        if radios_shape is None:
            n_z, n_a, n_x = get_2D_3D_shape(sinos_shape)
            radios_shape = (n_a, n_z, n_x)
        elif len(radios_shape) == 2:
            radios_shape = (1, ) + radios_shape

        self.sinos_shape = sinos_shape
        self.radios_shape = radios_shape
        n_a, n_z, n_x = radios_shape
        self.n_angles = n_a
        self.n_z = n_z
        self.n_x = n_x


    def set_rot_center(self, rot_center):
        """
        Set the rotation axis position for the current radios/sinos stack.

        rot_center: int or array
            Rotation axis position. A scalar indicates the same rotation axis position
            for all the projections.
        """
        if rot_center is None:
            rot_center = (self.n_x - 1) / 2.
        if not(np.isscalar(rot_center)):
            rot_center = np.array(rot_center)
            if rot_center.size != self.n_angles:
                raise ValueError(
                    "Expected rot_center to have %d elements but got %d"
                    % (self.n_angles, rot_center.size)
                )
        self.rot_center = rot_center
        self._rot_center_int = int(round(self.rot_center))

    def _configure_halftomo(self, halftomo, interpolate, angles):
        self.halftomo = halftomo
        self.interpolate = interpolate
        self.angles = angles
        self._halftomo_flip = False
        if not self.halftomo:
            return
        if interpolate and (angles is None):
            raise ValueError(
                "The parameter 'angles' has to be provided when using halftomo=True and interpolate=True"
            )
        # If CoR is on the left: "flip" the logic
        rc = self._rot_center_int
        if rc < (self.n_x - 1)/2:
            rc = self.n_x - 1 - rc
            self._rot_center_int = rc
            self.rot_center = self.n_x - self.rot_center
            self._halftomo_flip = True
        #
        if abs(self.rot_center - ((self.n_x - 1) / 2.)) < 1: # which tol ?
            raise ValueError(
                "Half tomography: incompatible rotation axis position: %.2f"
                % self.rot_center
            )
        self.sinos_halftomo_shape = (self.n_z, (self.n_angles + 1)// 2, 2 * self._rot_center_int)


    def _check_array_shape(self, array, kind="radio"):
        expected_shape = self.radios_shape if "radio" in kind else self.sinos_shape
        assert array.shape == expected_shape, "Expected radios shape %s, but got %s" % (expected_shape, array.shape)


    def _radios_to_sinos_simple(self, radios, output, copy=False):
        sinos = np.rollaxis(radios, 1, 0)  # view
        if not(copy) and output is None:
            return sinos
        if output is None: # copy and output is None
            return np.ascontiguousarray(sinos)  # copy
        # not(copy) and output is not None
        for i in range(output.shape[0]):
            output[i] = sinos[i]
        return output


    def _radios_to_sinos_halftomo(self, radios, sinos):
        n_a, n_z, n_x = radios.shape
        n_a2 = (n_a + 1) // 2
        out_dwidth = 2 * self._rot_center_int
        if sinos is not None:
            if sinos.shape[-1] != out_dwidth:
                raise ValueError(
                    "Expected sinos sinogram last dimension to have %d elements"
                    % out_dwidth
                )
            if sinos.shape[-2] != n_a2:
                raise ValueError("Expected sinograms to have %d angles" % n_a2)
        else:
            sinos = np.zeros(self.sinos_halftomo_shape, dtype=np.float32)
        for i in range(n_z):
            sino = radios[:, i, :]
            if self.interpolate:
                match_half_sinos_parts(sino, self.angles)
            elif n_a & 1:
                # Odd number of projections - add one line in the end
                sino = np.vstack([sino, np.zeros_like(sino[-1])])
            if self._halftomo_flip:
                sino = sino[:, ::-1]
            sinos[i][:] = convert_halftomo(sino, self._rot_center_int)
            if self._halftomo_flip:
                sinos[i][:] = sinos[i][:, ::-1]
        return sinos


    @property
    def output_shape(self):
        """
        Get the output sinograms shape.
        """
        if self.halftomo:
            return self.sinos_halftomo_shape
        return self.sinos_shape


    def radios_to_sinos(self, radios, output=None, copy=False):
        """
        Convert a chunk of radios to a stack of sinograms.

        Parameters
        -----------
        radios: array
            Radios to convert
        output: array, optional
            Output sinograms array, pre-allocated
        """
        self._check_array_shape(radios, kind="radio")
        if self.halftomo:
            return self._radios_to_sinos_halftomo(radios, output)
        return self._radios_to_sinos_simple(radios, output, copy=copy)

SinoProcessing = deprecated_class(
    "'SinoProcessing' was renamed 'SinoBuilder'", do_print=True
)(SinoBuilder)


def convert_halftomo(sino, rotation_axis_position, transition_width=None):
    """
    Converts a sinogram into a sinogram with extended FOV with the "half tomography"
    setting.
    """
    assert sino.ndim == 2
    assert (sino.shape[0] % 2) == 0

    na, nx = sino.shape
    if rotation_axis_position > nx:
        return _convert_halftomo_right(sino, rotation_axis_position)

    na2 = na // 2
    r = rotation_axis_position
    d = transition_width or nx - r
    res = np.zeros((na2, 2 * r), dtype="f")

    sino1 = sino[:na2, :]
    sino2 = sino[na2:, ::-1]
    res[:, : r - d] = sino1[:, : r - d]
    #
    w1 = np.linspace(0, 1, 2*d, endpoint=True)
    res[:, r - d:r + d] = (1 - w1) * sino1[:, r - d :] + w1 * sino2[:, 0 : 2 * d]
    #
    res[:, r+d:] = sino2[:, 2 * d :]

    return res


def match_half_sinos_parts(sino, angles, output=None):
    """
    Modifies the lower part of the half-acquisition sinogram so that each projection pair is
    separated by exactly 180 degrees.
    This means that `new_sino[k]` and `new_sino[k + n_angles//2]` will be 180 degrees apart.

    Parameters
    ----------
    sino: numpy.ndarray
        Two dimensional array with the sinogram in the form (n_angles, n_x)
    angles: numpy.ndarray
        One dimensional array with the rotation angles.
    output: numpy.array, optional
        Output sinogram. By default, the array 'sino' is modified in-place.

    Notes
    -----
    This function assumes that the angles are in an increasing order.
    """
    n_a = angles.size
    n_a_2 = n_a // 2
    sino_part1 = sino[:n_a_2, :]
    sino_part2 = sino[n_a_2:, :]
    angles = np.rad2deg(angles) # more numerically stable ?
    angles_1 = angles[:n_a_2]
    angles_2 = angles[n_a_2:]
    angles_2_target = angles_1 + 180.
    interpolator = interp1d(
        angles_2, sino_part2,
        axis=0, kind="linear", copy=False, fill_value="extrapolate"
    )
    if output is None:
        output = sino
    else:
        output[:n_a_2, :] = sino[:n_a_2, :]
    output[n_a_2:, :] = interpolator(angles_2_target)
    return output




# EXPERIMENTAL
def _convert_halftomo_right(sino, rotation_axis_position):
    """
    Converts a sinogram into a sinogram with extended FOV with the "half tomography"
    setting, with a CoR outside the image support.
    """
    assert sino.ndim == 2
    na, nx = sino.shape
    assert (na % 2) == 0
    assert rotation_axis_position > nx

    sino2 = np.pad(sino, ((0, 0), (0, rotation_axis_position - nx)), mode="reflect")
    return convert_halftomo(sino2, rotation_axis_position)


class SinoNormalization:
    """
    A class for sinogram normalization utilities.
    """

    kinds = [
        "chebyshev"
    ]

    def __init__(self, kind="chebyshev", sinos_shape=None, radios_shape=None):
        """
        Initialize a SinoNormalization class.

        Parameters
        -----------
        kind: str, optional
            Normalization type. They can be the following:
               - chebyshev: Each sinogram line is estimated by a Chebyshev polynomial
               of degree 2. This estimation is then subtracted from the sinogram.
            Default is "chebyshev"
        """
        self._get_shapes(sinos_shape, radios_shape)
        self._set_kind(kind)


    _get_shapes = SinoBuilder._get_shapes


    def _set_kind(self, kind):
        check_supported(kind, self.kinds, "sinogram normalization kind")
        self.normalization_kind = kind


    def _normalize_chebyshev_2D(self, sino):
        output = sino # inplace
        Nr, Nc = sino.shape
        J = np.arange(Nc)
        x = 2.* (J + 0.5 - Nc/2)/Nc
        sum0 = Nc
        f2 = (3.0*x*x-1.0)
        sum1 = (x**2).sum()
        sum2 = (f2**2).sum()
        for i in range(Nr):
            ff0 = sino[i, :].sum()
            ff1 = (x * sino[i, :]).sum()
            ff2 = (f2*sino[i, :]).sum()
            output[i, :] = sino[i, :] - (ff0/sum0 + ff1*x/sum1 + ff2*f2/sum2)
        return output


    def _normalize_chebyshev_3D(self, sino):
        for i in range(sino.shape[0]):
            self._normalize_chebyshev_2D(sino[i])
        return sino


    def _normalize_chebyshev(self, sino):
        if sino.ndim == 2:
            res = self._normalize_chebyshev_2D(sino)
        else:
            res = self._normalize_chebyshev_3D(sino)
        return res


    def _get_norm_func_name(self, prefix=""):
        return prefix + "normalize_" + self.normalization_kind


    def normalize(self, sino):
        """
        Normalize a sinogram or stack of sinogram.
        The process is done in-place, meaning that the sinogram content is overwritten.
        """
        norm_func = getattr(self, self._get_norm_func_name(prefix="_"))
        return norm_func(sino)




# -*- coding: utf-8 -*-
"""
Fourier filters.
"""
import numpy as np
import scipy.special as spspe

def get_lowpass_filter(img_shape, cutoff_par=None, use_rfft=False, data_type=np.float64):
    """Computes a low pass filter using the erfc function.

    Parameters
    ----------
    img_shape: tuple
        Shape of the image
    cutoff_par: float or sequence of two floats
        Position of the cut off in pixels, if a sequence is given the second float expresses the
        width of the transition region which is given as a fraction of the cutoff frequency.
        When only one float is given for this argument a gaussian is applied whose sigma is the
        parameter.
        When a sequence of two numbers is given then the filter is 1 ( no filtering) till the cutoff
        frequency while a smooth erfc transition to zero is done
    use_rfft: boolean, optional
        Creates a filter to be used with the result of a rfft type of Fourier transform. Defaults to False.
    data_type: `numpy.dtype`, optional
        Specifies the data type of the computed filter. It defaults to `numpy.float64`

    Raises
    ------
    ValueError
        In case of malformed cutoff_par

    Returns
    -------
    numpy.array_like
        The computed filter
    """
    if cutoff_par is None:
        return 1
    elif isinstance(cutoff_par, (int, float)):
        cutoff_pix = cutoff_par
        cutoff_trans_fact = None
    else:
        try:
            cutoff_pix, cutoff_trans_fact = cutoff_par
        except ValueError:
            raise ValueError(
                "Argument cutoff_par  (which specifies the pass filter shape) must be either a scalar or a"
                " sequence of two scalars"
            )
        if (not isinstance(cutoff_pix, (int, float))) or (not isinstance(cutoff_trans_fact, (int, float))):
            raise ValueError(
                "Argument cutoff_par  (which specifies the pass filter shape) must be  one number or a sequence"
                "of two numbers"
            )

    coords = [np.fft.fftfreq(s, 1) for s in img_shape]
    coords = np.meshgrid(*coords, indexing="ij")

    r = np.sqrt(np.sum(np.array(coords, dtype=data_type) ** 2, axis=0))

    if cutoff_trans_fact is not None:
        k_cut = 0.5 / cutoff_pix
        k_cut_width = k_cut * cutoff_trans_fact
        k_pos_rescaled = (r - k_cut) / k_cut_width
        res = spspe.erfc(k_pos_rescaled) / 2
    else:
        res = np.exp(-(np.pi ** 2) * (r ** 2) * (cutoff_pix ** 2) * 2)

    # Making sure to force result to chosen data type
    res = res.astype(data_type)

    if use_rfft:
        slicelist = [slice(None)] * (len(res.shape) - 1) + [slice(0, res.shape[-1] // 2 + 1)]
        return res[tuple(slicelist)]
    else:
        return res


def get_highpass_filter(img_shape, cutoff_par=None, use_rfft=False, data_type=np.float64):
    """Computes a high pass filter using the erfc function.

    Parameters
    ----------
    img_shape: tuple
        Shape of the image
    cutoff_par: float or sequence of two floats
        Position of the cut off in pixels, if a sequence is given the second float expresses the
        width of the transition region which is given as a fraction of the cutoff frequency.
        When only one float is given for this argument a gaussian is applied whose sigma is the
        parameter, and the result is subtracted from 1 to obtain the high pass filter
        When a sequence of two numbers is given then the filter is 1 ( no filtering) above the cutoff
        frequency and then a smooth  transition to zero is done for smaller frequency
    use_rfft: boolean, optional
        Creates a filter to be used with the result of a rfft type of Fourier transform. Defaults to False.
    data_type: `numpy.dtype`, optional
        Specifies the data type of the computed filter. It defaults to `numpy.float64`

    Raises
    ------
    ValueError
        In case of malformed cutoff_par

    Returns
    -------
    numpy.array_like
        The computed filter
    """
    if cutoff_par is None:
        return 1
    else:
        return 1 - get_lowpass_filter(img_shape, cutoff_par, use_rfft=use_rfft, data_type=data_type)


def get_bandpass_filter(img_shape, cutoff_lowpass=None, cutoff_highpass=None, use_rfft=False, data_type=np.float64):
    """Computes a band pass filter using the erfc function.

    The cutoff structures should be formed as follows:

    - tuple of two floats: the first indicates the cutoff frequency, the second \
        determines the width of the transition region, as fraction of the cutoff frequency.

    - one float -> it represents the sigma of a gaussian which acts as a filter or anti-filter (1 - filter).

    Parameters
    ----------
    img_shape: tuple
        Shape of the image
    cutoff_lowpass: float or sequence of two floats
        Cutoff parameters for the low-pass filter
    cutoff_highpass: float or sequence of two floats
        Cutoff parameters for the high-pass filter
    use_rfft: boolean, optional
        Creates a filter to be used with the result of a rfft type of Fourier transform. Defaults to False.
    data_type: `numpy.dtype`, optional
        Specifies the data type of the computed filter. It defaults to `numpy.float64`

    Raises
    ------
    ValueError
        In case of malformed cutoff_par

    Returns
    -------
    numpy.array_like
        The computed filter
    """
    return get_lowpass_filter(
        img_shape, cutoff_par=cutoff_lowpass, use_rfft=use_rfft, data_type=data_type
    ) * get_highpass_filter(img_shape, cutoff_par=cutoff_highpass, use_rfft=use_rfft, data_type=data_type)



from ..pipeline.fullfield.chunked import ChunkedPipeline as FullFieldPipeline
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.resources.app.fullfield was moved to nabu.pipeline.fullfield.chunked.",
    func_name="fullfield_pipeline"
)


from ast import literal_eval
import numpy as np
from psutil import virtual_memory, cpu_count

def get_values_from_file(fname, n_values=None, shape=None, sep=None, any_size=False):
    """
    Read a text file and scan the values inside.
    This function expects one value per line, or values separated with a separator
    defined with the `sep` parameter.

    Parameters
    ----------
    fname: str
        Path of the text file
    n_values: int, optional
        If set to a value, this function will check that it scans exactly this
        number of values.
        Ignored if `shape` is provided
    shape: tuple, optional
        Generalization of n_values for higher dimensions.
    sep: str, optional
        Separator between values. Default is white space
    any_size: bool, optional
        If set to True, then the parameters 'n_values' and 'shape' are ignored.

    Returns
    --------
    arr: numpy.ndarray
        An array containing the values scanned from the text file
    """
    if not(any_size) and not((n_values is not None) ^ (shape is not None)):
        raise ValueError("Please provide either n_values or shape")
    arr = np.loadtxt(fname, ndmin=1)
    if (n_values is not None) and (arr.shape[0] != n_values):
        raise ValueError("Expected %d values, but could get %d values" % (n_values, arr.shape[0]))
    if (shape is not None) and (arr.shape != shape):
        raise ValueError("Expected shape %s, but got shape %s" % (shape, arr.shape))
    return arr


def get_memory_per_node(max_mem, is_percentage=True):
    """
    Get the available memory per node in GB.

    Parameters
    ----------
    max_mem: float
        If is_percentage is False, then number is interpreted as an absolute
        number in GigaBytes.
        Otherwise, it should be a number between 0 and 100 and is interpreted
        as a percentage.
    is_percentage: bool
        A boolean indicating whether the parameter max_mem is to be interpreted
        as a percentage of available system memory.
    """
    sys_avail_mem = virtual_memory().available / 1e9
    if is_percentage:
        return (max_mem/100.) * sys_avail_mem
    else:
        return min(max_mem, sys_avail_mem)


def get_threads_per_node(max_threads, is_percentage=True):
    """
    Get the available memory per node in GB.

    Parameters
    ----------
    max_threads: float
        If is_percentage is False, then number is interpreted as an absolute
        number of threads.
        Otherwise, it should be a number between 0 and 100 and is interpreted
        as a percentage.
    is_percentage: bool
        A boolean indicating whether the parameter max_threads is to be interpreted
        as a percentage of available system memory.
    """
    sys_n_threads = cpu_count(logical=True)
    if is_percentage:
        return (max_threads/100.) * sys_n_threads
    else:
        return min(max_threads, sys_n_threads)


def extract_parameters(params_str, sep=";"):
    """
    Extract the named parameters from a string.

    Example
    --------
    The function can be used as follows:

    >>> extract_parameters("window_width=None; median_filt_shape=(3,3); padding_mode='wrap'")
    ... {'window_width': None, 'median_filt_shape': (3, 3), 'padding_mode': 'wrap'}
    """
    if params_str in ("", None):
        return {}
    params_list = params_str.strip(sep).split(sep)
    res = {}
    for param_str in params_list:
        param_name, param_val_str = param_str.strip().split("=")
        param_name = param_name.strip()
        param_val_str = param_val_str.strip()
        param_val = literal_eval(param_val_str)
        res[param_name] = param_val
    return res


def is_hdf5_extension(ext, errors="ignore"):
    if ext.lower() in ["h5", "hdf5", "nx"]:
        return True
    else:
        if errors == "raise":
            raise ValueError("Unknown data file format: %s" % ext)
        return False



import numpy as np
from pycuda import gpuarray as garray
from ..utils import calc_padding_lengths, updiv, get_cuda_srcfile
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from ..cuda.padding import CudaPadding
from .phase_cuda import CudaPaganinPhaseRetrieval
from .ctf import CTFPhaseRetrieval


# TODO:
#  - better padding scheme (for now 2*shape)
#  - rework inheritance scheme ? (base class SingleDistancePhaseRetrieval and its cuda counterpart)
class CudaCTFPhaseRetrieval(CTFPhaseRetrieval):

    """
    Cuda back-end of CTFPhaseRetrieval
    """

    def __init__(
        self, shape, geo_pars, delta_beta,
        padding_mode="reflect", lim1=1.0e-5, lim2=0.2, use_rfft=True,
        logger=None, cuda_options=None,
    ):
        """
        Initialize a CudaCTFPhaseRetrieval.

        Parameters
        ----------
        shape: tuple
            Shape of the images to process
        padding_mode: str
            Padding mode. Default is "reflect".

        Other parameters
        -----------------
        Please refer to CTFPhaseRetrieval documentation.
        """
        if not use_rfft:
            raise ValueError("Only use_rfft=True is supported")
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        super().__init__(geo_pars, delta_beta, lim1=lim1, lim2=lim2, logger=logger, use_rfft=True)
        self._calc_shape(shape)
        self._init_ctf_filter()
        self._init_padding(padding_mode)
        self._init_fft()
        self._init_mult_kernel()


    def _calc_shape(self, shape):
        if np.isscalar(shape):
            shape = (shape, shape)
        else:
            assert len(shape) == 2
        self.shape = shape
        n_y0, n_x0 = self.shape
        n_y_p = 2 * n_y0
        n_x_p = 2 * n_x0
        self.shape_padded = (n_y_p, n_x_p)


    def _init_ctf_filter(self):
        self.require_filter_for_shape(self.shape_padded)
        self._mean_scale_factor = self.unreg_filter_denom[0, 0] * np.prod(self.shape_padded)
        self._d_filter_num = garray.to_gpu(self.unreg_filter_denom).astype("f")
        self._d_filter_denom = garray.to_gpu(
            (1. / (2 * self.unreg_filter_denom * self.unreg_filter_denom + self.lim)).astype("f")
        )


    def _init_padding(self, padding_mode):
        self.padding_mode = padding_mode
        pad_width = calc_padding_lengths(self.shape, self.shape_padded)
        # Custom coordinate transform to get directly FFT layout
        R, C = np.indices(self.shape, dtype=np.int32)
        coords_R = np.roll(
            np.pad(R, pad_width, mode=padding_mode),
            (-pad_width[0][0], -pad_width[1][0]),
            axis=(0, 1)
        )
        coords_C = np.roll(
            np.pad(C, pad_width, mode=padding_mode),
            (-pad_width[0][0], -pad_width[1][0]),
            axis=(0, 1)
        )
        self.cuda_padding = CudaPadding(
            self.shape,
            (coords_R, coords_C),
            mode=self.padding_mode,
            # propagate cuda options ?
        )


    def _init_fft(self):
        # Import has to be done here, otherwise scikit-cuda creates a cuda/cublas context at import
        from silx.math.fft.cufft import CUFFT
        self.cufft = CUFFT(template=np.zeros(self.shape_padded, dtype="f"))
        self.d_radio_padded = self.cufft.data_in
        self.d_radio_f = self.cufft.data_out


    def _init_mult_kernel(self):
        self.cpxmult_kernel = CudaKernel(
            "CTF_kernel",
            filename=get_cuda_srcfile("ElementOp.cu"),
            signature="PPPfii",
        )
        Nx = np.int32(self.shape_padded[1]//2 + 1)
        Ny = np.int32(self.shape_padded[0])
        self._cpxmult_kernel_args = [
            self.d_radio_f,
            self._d_filter_num,
            self._d_filter_denom,
            np.float32(self._mean_scale_factor),
            Nx,
            Ny
        ]
        blk = (32, 32, 1)
        grd = (updiv(Nx, blk[0]), updiv(Ny, blk[1]))
        self._cpxmult_kernel_kwargs = {"grid": grd, "block": blk}


    set_input = CudaPaganinPhaseRetrieval.set_input


    def retrieve_phase(self, image, output=None, normalize_by_mean=False):
        """
        Perform padding on an image. Please see the documentation of CTFPhaseRetrieval.retrieve_phase().
        """
        self.set_input(image)
        self.cuda_padding.pad(image, output=self.d_radio_padded)
        if normalize_by_mean:
            m = garray.sum(self.d_radio_padded).get() / np.prod(self.shape_padded)
            self.d_radio_padded /= m
        self.cufft.fft(self.d_radio_padded, output=self.d_radio_f)
        self.cpxmult_kernel(*self._cpxmult_kernel_args, **self._cpxmult_kernel_kwargs)
        self.cufft.ifft(self.d_radio_f, output=self.d_radio_padded)

        if output is None:
            output = self.cuda_processing.allocate_array("d_output", self.shape)
        output[:, :] = self.d_radio_padded[:self.shape[0], :self.shape[1]]
        return output



import numpy as np
from scipy.ndimage import convolve1d
from silx.image.utils import gaussian_kernel

class UnsharpMask:
    """
    A helper class for unsharp masking.
    """

    avail_methods = ["gaussian", "log"]

    def __init__(self, shape, sigma, coeff, mode="reflect", method="gaussian"):
        """
        Initialize a Unsharp mask.
        `UnsharpedImage =  (1 + coeff)*Image - coeff * ConvolutedImage`

        If method == "log":
        `UnsharpedImage = Image + coeff*ConvolutedImage`

        Parameters
        -----------
        shape: tuple
            Shape of the image.
        sigma: float
            Standard deviation of the Gaussian kernel
        coeff: float
            Coefficient in the linear combination of unsharp mask
        mode: str, optional
            Convolution mode. Default is "reflect"
        method: str, optional
            Method of unsharp mask. Can be "gaussian" (default) or "log" for
            Laplacian of Gaussian.
        """
        self.shape = shape
        self.ndim = len(self.shape)
        self.sigma = sigma
        self.coeff = coeff
        self._set_method(method)
        self.mode = mode
        self._compute_gaussian_kernel()

    def _set_method(self, method):
        if method not in self.avail_methods:
            raise ValueError(
                "Unknown unsharp method '%s'. Available are %s"
                % (method, str(self.avail_methods))
            )
        self.method = method

    def _compute_gaussian_kernel(self):
        self._gaussian_kernel = np.ascontiguousarray(gaussian_kernel(self.sigma), dtype=np.float32)

    def _blur2d(self, image):
        res1 = convolve1d(image, self._gaussian_kernel, axis=1, mode=self.mode)
        res = convolve1d(res1, self._gaussian_kernel, axis=0, mode=self.mode)
        return res

    def unsharp(self, image, output=None):
        """
        Reference unsharp mask implementation.
        """
        image_b = self._blur2d(image)
        if self.method == "gaussian":
            res = (1 + self.coeff) * image - self.coeff * image_b
        else: # LoG
            res = image + self.coeff * image_b
        if output is not None:
            output[:] = res[:]
            return output
        return res
